#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, cast, Dict, List, Optional, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot.enums import enum_to_list, PromptType, VectorDatabaseRetrievers
from datarobot.models.api_object import APIObject
from datarobot.models.custom_model import CustomModelVersion
from datarobot.models.genai.llm import LLMDefinition
from datarobot.models.genai.playground import Playground
from datarobot.models.genai.vector_database import VectorDatabase
from datarobot.utils import to_api
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


def get_entity_id(
    entity: Union[Playground, LLMDefinition, VectorDatabase, LLMBlueprint, str]
) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be entity ID or the entity.

    Returns
    -------
    id : str
        Entity ID
    """
    if isinstance(entity, str):
        return entity

    return entity.id


class VectorDatabaseSettingsDict(TypedDict):
    max_documents_retrieved_per_prompt: Optional[int]
    max_tokens: Optional[int]
    retriever: Optional[VectorDatabaseRetrievers]
    add_neighbor_chunks: Optional[bool]


class LLMSettingsCommonDict(TypedDict):
    system_prompt: Optional[str]
    max_completion_length: Optional[int]
    temperature: Optional[float]
    top_p: Optional[float]


class LLMSettingsCustomModelDict(TypedDict):
    system_prompt: Optional[str]
    validation_id: Optional[str]


vector_database_settings_trafaret = t.Dict(
    {
        t.Key("max_documents_retrieved_per_prompt", optional=True): t.Or(t.Int, t.Null),
        t.Key("max_tokens", optional=True): t.Or(t.Int, t.Null),
        t.Key("retriever", optional=True): t.Enum(*VectorDatabaseRetrievers._member_names_),
        t.Key("add_neighbor_chunks", optional=True): t.Bool,
    }
).ignore_extra("*")

llm_blueprint_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
        t.Key("description"): t.String(allow_blank=True),
        t.Key("is_saved"): t.Bool,
        t.Key("is_starred"): t.Bool,
        t.Key("playground_id"): t.String,
        t.Key("llm_id", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_settings", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        t.Key("llm_name", optional=True): t.Or(t.String(allow_blank=True), t.Null),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("creation_user_name"): t.String(allow_blank=True),
        t.Key("last_update_date"): t.String,
        t.Key("last_update_user_id"): t.String,
        t.Key("prompt_type"): t.Enum(*enum_to_list(PromptType)),
        t.Key("vector_database_id", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_settings", optional=True): t.Or(
            vector_database_settings_trafaret, t.Null
        ),
        t.Key("vector_database_name", optional=True): t.Or(t.String(allow_blank=True), t.Null),
        t.Key("vector_database_status", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_error_message", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_error_resolution", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_llm_validation_status", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_llm_error_message", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_llm_error_resolution", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


class VectorDatabaseSettings(APIObject):
    """
    Settings for a DataRobot GenAI vector database associated with an LLM blueprint.

    Attributes
    ----------
    max_documents_retrieved_per_prompt : int or None, optional
        The maximum number of documents to retrieve for each prompt.
    max_tokens : int or None, optional
        The maximum number of tokens to retrieve for each document.
    retriever: VectorDatabaseRetrievers
        The vector database retriever name.
    add_neighbor_chunks
        Whether to add neighboring documents to the retrieved documents.

    """

    _converter = vector_database_settings_trafaret

    def __init__(
        self,
        max_documents_retrieved_per_prompt: Optional[int] = None,
        max_tokens: Optional[int] = None,
        retriever: Optional[
            VectorDatabaseRetrievers
        ] = VectorDatabaseRetrievers.SINGLE_LOOKUP_RETRIEVER,
        add_neighbor_chunks: Optional[bool] = False,
    ):
        self.max_documents_retrieved_per_prompt = max_documents_retrieved_per_prompt
        self.max_tokens = max_tokens
        self.max_documents_retrieved_per_prompt = max_documents_retrieved_per_prompt
        self.max_tokens = max_tokens
        self.retriever = retriever
        self.add_neighbor_chunks = add_neighbor_chunks

    def to_dict(self) -> VectorDatabaseSettingsDict:
        return {
            "max_documents_retrieved_per_prompt": self.max_documents_retrieved_per_prompt,
            "max_tokens": self.max_tokens,
            "retriever": self.retriever,
            "add_neighbor_chunks": self.add_neighbor_chunks,
        }


class LLMBlueprint(APIObject):
    """
    Metadata for a DataRobot GenAI LLM blueprint.

    Attributes
    ----------
    id : str
        The LLM blueprint ID.
    name : str
        The LLM blueprint name.
    description : str
        A description of the LLM blueprint.
    is_saved : bool
        Whether the LLM blueprint is saved (meaning the settings are locked
        and blueprint is eligible for use with `ComparisonPrompts`).
    is_starred : bool
        Whether the LLM blueprint is starred.
    playground_id : str
        The ID of the playground associated with the LLM blueprint.
    llm_id : str or None
        The ID of the LLM type. If not None, this must be one of the IDs returned by `LLMDefinition.list`
        for this user.
    llm_name : str or None
        The name of the LLM.
    llm_settings : dict or None
        The LLM settings for the LLM blueprint. The specific keys allowed and the
        constraints on the values are defined in the response from `LLMDefinition.list`
        but this typically has dict fields:
        - system_prompt - The system prompt that tells the LLM how to behave.
        - max_completion_length - The maximum number of tokens in the completion.
        - temperature - Controls the variability in the LLM response.
        - top_p - Whether the model considers next tokens with top_p probability mass.
        Or
        - system_prompt - The system prompt that tells the LLM how to behave.
        - validation_id - The ID of the external model LLM validation.
        - external_llm_context_size - The external LLM's context size, in tokens,
        for external model LLM blueprints.
    creation_date : str
        The date the playground was created.
    creation_user_id : str
        The ID of the user creating the playground.
    creation_user_name : str
        The name of the user creating the playground.
    last_update_date : str
        The date the playground was last updated.
    last_update_user_id : str
        The ID of the user who most recently updated the playground.
    prompt_type : PromptType
        The prompting strategy for the LLM blueprint.
        Currently supported options are listed in PromptType.
    vector_database_id : str or None
        The ID of the vector database, if any, associated with the LLM blueprint.
    vector_database_settings : VectorDatabaseSettings or None
        The settings for the vector database, if any, associated with the LLM blueprint.
    vector_database_name : str or None
        The name of the vector database associated with the LLM blueprint, if any.
    vector_database_status : str or None
        The status of the vector database, if any, associated with the LLM blueprint.
    vector_database_error_message : str or None
        The error message for the vector database, if any, associated with the LLM blueprint.
    vector_database_error_resolution : str or None
        The resolution for the vector database error, if any, associated with the LLM blueprint.
    custom_model_llm_validation_status : str or None
        The status of the custom model LLM validation if the llm_id is 'custom-model'.
    custom_model_llm_error_message : str or None
        The error message for the custom model LLM, if any.
    custom_model_llm_error_resolution : str or None
        The resolution for the custom model LLM error, if any.
    """

    _path = "api/v2/genai/llmBlueprints"

    _converter = llm_blueprint_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        is_saved: bool,
        is_starred: bool,
        playground_id: str,
        creation_date: str,
        creation_user_id: str,
        creation_user_name: str,
        last_update_date: str,
        last_update_user_id: str,
        prompt_type: PromptType,
        llm_id: Optional[str] = None,
        llm_name: Optional[str] = None,
        llm_settings: Optional[Union[LLMSettingsCommonDict, LLMSettingsCustomModelDict]] = None,
        vector_database_id: Optional[str] = None,
        vector_database_settings: Optional[Dict[str, Any]] = None,
        vector_database_name: Optional[str] = None,
        vector_database_status: Optional[str] = None,
        vector_database_error_message: Optional[str] = None,
        vector_database_error_resolution: Optional[str] = None,
        custom_model_llm_validation_status: Optional[str] = None,
        custom_model_llm_error_message: Optional[str] = None,
        custom_model_llm_error_resolution: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.is_saved = is_saved
        self.is_starred = is_starred
        self.playground_id = playground_id
        self.llm_id = llm_id
        self.llm_name = llm_name
        self.llm_settings = llm_settings
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.creation_user_name = creation_user_name
        self.last_update_date = last_update_date
        self.last_update_user_id = last_update_user_id
        self.prompt_type = prompt_type
        self.vector_database_id = vector_database_id
        self.vector_database_name = vector_database_name
        self.vector_database_status = vector_database_status
        self.vector_database_error_message = vector_database_error_message
        self.vector_database_error_resolution = vector_database_error_resolution
        self.custom_model_llm_validation_status = custom_model_llm_validation_status
        self.custom_model_llm_error_message = custom_model_llm_error_message
        self.custom_model_llm_error_resolution = custom_model_llm_error_resolution
        self.vector_database_settings = (
            VectorDatabaseSettings.from_server_data(vector_database_settings)
            if vector_database_settings
            else None
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id}, name={self.name}, is_saved={self.is_saved})"
        )

    @classmethod
    def create(
        cls,
        playground: Union[Playground, str],
        name: str,
        prompt_type: PromptType = PromptType.CHAT_HISTORY_AWARE,
        description: str = "",
        llm: Optional[Union[LLMDefinition, str]] = None,
        llm_settings: Optional[Union[LLMSettingsCommonDict, LLMSettingsCustomModelDict]] = None,
        vector_database: Optional[Union[VectorDatabase, str]] = None,
        vector_database_settings: Optional[VectorDatabaseSettings] = None,
    ) -> LLMBlueprint:
        """
        Create a new LLM blueprint.

        Parameters
        ----------
        playground : Playground or str
            The playground associated with the created LLM blueprint.
            Accepts playground or playground ID.
        name : str
            The LLM blueprint name.
        prompt_type : PromptType, optional
            Prompting type of the LLM blueprint, by default PromptType.CHAT_HISTORY_AWARE.
        description : Optional[str]
            An optional description for the LLM blueprint, otherwise null.
        llm : LLMDefinition, str, or None, optional
            The LLM to use for the blueprint, either `LLMDefinition` or LLM ID.
        llm_settings : dict or None
            The LLM settings for the LLM blueprint. The specific keys allowed and the
            constraints on the values are defined in the response from `LLMDefinition.list`
            but this typically has dict fields:
            - system_prompt - The system prompt that tells the LLM how to behave.
            - max_completion_length - The maximum number of tokens in the completion.
            - temperature - Controls the variability in the LLM response.
            - top_p - Whether the model considers next tokens with top_p probability mass.
            Or
            - system_prompt - The system prompt that tells the LLM how to behave.
            - validation_id - The ID of the custom model LLM validation
            for custom model LLM blueprints.
        vector_database: VectorDatabase, str, or None, optional
            The vector database to use with this LLM blueprint, either
            `VectorDatabase` or vector database ID.
        vector_database_settings: VectorDatabaseSettings or None, optional
            The settings for the vector database, if any.

        Returns
        -------
        llm_blueprint : LLMBlueprint
            The created LLM blueprint.
        """
        payload = {
            "playground_id": get_entity_id(playground),
            "name": name,
            "prompt_type": prompt_type,
            "description": description,
            "llm_id": get_entity_id(llm) if llm else None,
            "llm_settings": llm_settings,
            "vector_database_id": get_entity_id(vector_database) if vector_database else None,
            "vector_database_settings": (
                vector_database_settings.to_dict() if vector_database_settings else None
            ),
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def create_from_llm_blueprint(
        cls,
        llm_blueprint: Union[LLMBlueprint, str],
        name: str,
        description: str = "",
    ) -> LLMBlueprint:
        """
        Create a new LLM blueprint from an existing LLM blueprint.

        Parameters
        ----------
        llm_blueprint : LLMBlueprint or str
            The LLM blueprint to use to create the new LLM blueprint.
            Accepts LLM blueprint or LLM blueprint ID.
        name : str
            LLM blueprint name.
        description : Optional[str]
            Description of the LLM blueprint, by default "".

        Returns
        -------
        llm_blueprint : LLMBlueprint
            The created LLM blueprint.
        """
        payload = {
            "llm_blueprint_id": get_entity_id(llm_blueprint),
            "name": name,
            "description": description,
        }

        url = f"{cls._client.domain}/{cls._path}/fromLLMBlueprint/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, llm_blueprint_id: str) -> LLMBlueprint:
        """
        Retrieve a single LLM blueprint.

        Parameters
        ----------
        llm_blueprint_id : str
            The ID of the LLM blueprint you want to retrieve.

        Returns
        -------
        llm_blueprint : LLMBlueprint
            The requested LLM blueprint.
        """
        url = f"{cls._client.domain}/{cls._path}/{llm_blueprint_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        playground: Optional[Union[Playground, str]] = None,
        llms: Optional[List[Union[LLMDefinition, str]]] = None,
        vector_databases: Optional[List[Union[VectorDatabase, str]]] = None,
        is_saved: Optional[bool] = None,
        is_starred: Optional[bool] = None,
        sort: Optional[str] = None,
    ) -> List[LLMBlueprint]:
        """
        Lists all LLM blueprints available to the user. If the playground is specified, then the
        results are restricted to the LLM blueprints associated with the playground. If the
        LLMs are specified, then the results are restricted to the LLM blueprints using those
        LLM types. If `vector_databases` are specified, then the results are restricted to the
        LLM blueprints using those vector databases.

        Parameters
        ----------
        playground : Optional[Union[Playground, str]], optional
            The returned LLM blueprints are filtered to those associated with a specific playground
            if it is specified. Accepts either the entity or the ID.
        llms : Optional[list[Union[LLMDefinition, str]]], optional
            The returned LLM blueprints are filtered to those associated with the LLM types
            specified. Accepts either the entity or the ID.
        vector_databases : Optional[list[Union[VectorDatabase, str]]], optional
            The returned LLM blueprints are filtered to those associated with the vector databases
            specified. Accepts either the entity or the ID.
        is_saved: Optional[bool], optional
            The returned LLM blueprints are filtered to those matching is_saved.
        is_starred: Optional[bool], optional
            The returned LLM blueprints are filtered to those matching is_starred.
        sort : Optional[str]
            Property to sort LLM blueprints by.
            Prefix the attribute name with a dash to sort in descending order,
            e.g. sort='-creationDate'.
            Currently supported options are listed in ListLLMBlueprintsSortQueryParams
            but the values can differ with different platform versions.
            By default, the sort parameter is None which will result in
            LLM blueprints being returned in order of creation time descending.

        Returns
        -------
        playgrounds : list[Playground]
            A list of playgrounds available to the user.
        """
        params = {
            "playground_id": get_entity_id(playground) if playground else None,
            "llm_ids": [get_entity_id(llm) for llm in llms] if llms else None,
            "vector_databases": (
                [get_entity_id(vdb) for vdb in vector_databases] if vector_databases else None
            ),
            "is_saved": is_saved,
            "is_starred": is_starred,
            "sort": sort,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        llm: Optional[Union[LLMDefinition, str]] = None,
        llm_settings: Optional[Union[LLMSettingsCommonDict, LLMSettingsCustomModelDict]] = None,
        vector_database: Optional[Union[VectorDatabase, str]] = None,
        vector_database_settings: Optional[VectorDatabaseSettings] = None,
        is_saved: Optional[bool] = None,
        is_starred: Optional[bool] = None,
        prompt_type: Optional[PromptType] = None,
        remove_vector_database: Optional[bool] = False,
    ) -> LLMBlueprint:
        """
        Update the LLM blueprint.

        Parameters
        ----------
        name : str or None, optional
            The new name for the LLM blueprint.
        description: str or None, optional
            The new description for the LLM blueprint.
        llm: Optional[Union[LLMDefinition, str]], optional
            The new LLM type for the LLM blueprint.
        llm_settings: Optional[dict], optional
            The new LLM settings for the LLM blueprint. These must match the `LLMSettings`
            returned from the `LLMDefinition.list` method for the LLM type used for this
            LLM blueprint but this typically has dict fields:
            - system_prompt - The system prompt that tells the LLM how to behave.
            - max_completion_length - The maximum number of tokens in the completion.
            - temperature - Controls the variability in the LLM response.
            - top_p - Whether the model considers next tokens with top_p probability mass.
            Or
            - system_prompt - The system prompt that tells the LLM how to behave.
            - validation_id - The ID of the custom model LLM validation
            for custom model LLM blueprints.
        vector_database: Optional[Union[VectorDatabase, str]], optional
            The new vector database for the LLM blueprint.
        vector_database_settings: Optional[VectorDatabaseSettings], optional
            The new vector database settings for the LLM blueprint.
        is_saved: Optional[bool], optional
            The new `is_saved` attribute for the LLM blueprint (meaning the settings
            are locked and blueprint is eligible for use with `ComparisonPrompts`).
        is_starred: Optional[bool], optional
            The new setting for whether the LLM blueprint is starred.
        prompt_type : PromptType, optional
            The new prompting type of the LLM blueprint.
        remove_vector_database: Optional[bool], optional
            Whether to remove the vector database from the LLM blueprint.

        Returns
        -------
        llm_blueprint : LLMBlueprint
            The updated LLM blueprint.
        """
        payload = {
            "name": name,
            "description": description,
            "llm_id": get_entity_id(llm) if llm else None,
            "llm_settings": llm_settings,
            "vector_database_id": get_entity_id(vector_database) if vector_database else None,
            "vector_database_settings": (
                vector_database_settings.to_dict() if vector_database_settings else None
            ),
            "is_saved": is_saved,
            "is_starred": is_starred,
            "prompt_type": prompt_type,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        json_payload = cast(Dict[str, Any], to_api(payload))
        if remove_vector_database:  # This forces the removal of the vector database.
            json_payload["vectorDatabaseId"] = None
            json_payload["vectorDatabaseSettings"] = None
        r_data = self._client.patch(url, json=json_payload)
        return self.from_server_data(r_data.json())

    def delete(self) -> None:
        """
        Delete the single LLM blueprint.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    def register_custom_model(
        self,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        llm_test_configuration_ids: Optional[List[str]] = None,
    ) -> CustomModelVersion:
        """
        Create a new CustomModelVersion. This registers a custom model from the LLM blueprint.

        Parameters
        ----------
        prompt_column_name : Optional[str]
            The column name of the prompt text.
        target_column_name : Optional[str]
            The column name of the response text.
        llm_test_configuration_ids : List[str], optional
            The IDs of the LLM test configurations to execute.

        Returns
        -------
        custom_model : CustomModelVersion
            The registered custom model.
        """
        payload = {
            "llm_blueprint_id": self.id,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "llm_test_configuration_ids": llm_test_configuration_ids,
        }

        url = f"{self._client.domain}/api/v2/genai/customModelVersions/"
        r_data = self._client.post(url, data=payload)
        location = wait_for_async_resolution(self._client, r_data.headers["Location"])
        return CustomModelVersion.from_location(location)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Optional, Union

import trafaret as t

from datarobot.models.deployment import Deployment
from datarobot.models.genai.custom_model_validation import (
    custom_model_validation_trafaret,
    CustomModelValidation,
    get_entity_id,
)
from datarobot.models.model import Model
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id
from datarobot.utils.waiters import wait_for_async_resolution

custom_model_llm_validation_trafaret = (
    t.Dict(
        {
            t.Key("chat_model_id", optional=True, default=None): t.Or(t.Null, t.String),
        }
    )
    .merge(custom_model_validation_trafaret)
    .ignore_extra("*")
)


class CustomModelLLMValidation(CustomModelValidation):
    """
    Validation record checking the ability of the deployment to serve
    as a custom model LLM.

    Attributes
    ----------
    id : str
        The ID of the validation.
    prompt_column_name : str
        The name of the column the deployed model uses for prompt text input.
    target_column_name : str
        The name of the column the deployed model uses for prediction output.
    chat_model_id : Optional[str]
        The model ID to specify when calling the chat completion API of the deployment.
    deployment_id : str
        The ID of the deployment.
    model_id : str
        The ID of the underlying deployed model, which can be found using `Deployment.model["id"]`.
    validation_status : str
        Can be TESTING, FAILED, or PASSED. Only PASSED is allowed for use.
    deployment_access_data : dict, optional
        The data that will be used for accessing the deployment prediction server.
        This field is only available for deployments that pass validation.
        Dict fields are as follows:
        - prediction_api_url - The URL for the deployment prediction server.
        - datarobot_key - The first of two auth headers for the prediction server.
        - authorization_header - The second of two auth headers for the prediction server.
        - input_type - The input type the model expects, either JSON or CSV.
        - model_type - The target type of the deployed custom model.
    tenant_id : str
        The creating user's tenant ID.
    name : str
        The display name of the validated custom model.
    creation_date : str
        The creation date of the validation (ISO 8601 formatted).
    user_id : str
        The ID of the creating user.
    error_message : Optional[str]
        Additional information for the errored validation.
    deployment_name : Optional[str]
        The name of the validated deployment.
    user_name : Optional[str]
        The name of the creating user.
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    prediction_timeout: int
        The timeout in seconds for the prediction API used in this custom model validation.
    """

    _path = "api/v2/genai/customModelLLMValidations"
    _converter = custom_model_llm_validation_trafaret

    def __init__(
        self,
        id: str,
        prompt_column_name: str,
        target_column_name: str,
        chat_model_id: Optional[str],
        deployment_id: str,
        model_id: str,
        validation_status: str,
        deployment_access_data: Optional[Dict[str, Any]],
        tenant_id: str,
        name: str,
        creation_date: str,
        user_id: str,
        error_message: Optional[str],
        deployment_name: Optional[str],
        user_name: Optional[str],
        use_case_id: Optional[str],
        prediction_timeout: int,
    ):
        super().__init__(
            id=id,
            prompt_column_name=prompt_column_name,
            target_column_name=target_column_name,
            deployment_id=deployment_id,
            model_id=model_id,
            validation_status=validation_status,
            deployment_access_data=deployment_access_data,
            tenant_id=tenant_id,
            name=name,
            creation_date=creation_date,
            user_id=user_id,
            error_message=error_message,
            deployment_name=deployment_name,
            user_name=user_name,
            use_case_id=use_case_id,
            prediction_timeout=prediction_timeout,
        )
        self.id = id
        self.prompt_column_name = prompt_column_name
        self.target_column_name = target_column_name
        self.chat_model_id = chat_model_id
        self.deployment_id = deployment_id
        self.model_id = model_id
        self.validation_status = validation_status
        self.deployment_access_data = deployment_access_data
        self.tenant_id = tenant_id
        self.error_message = error_message
        self.name = name
        self.creation_date = creation_date
        self.user_id = user_id
        self.deployment_name = deployment_name
        self.user_name = user_name
        self.use_case_id = use_case_id
        self.prediction_timeout = prediction_timeout

    @classmethod
    def create(
        cls,
        deployment_id: Union[Deployment, str],
        model: Optional[Union[Model, str]] = None,
        use_case: Optional[Union[UseCase, str]] = None,
        name: Optional[str] = None,
        wait_for_completion: bool = False,
        prediction_timeout: Optional[int] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        chat_model_id: Optional[str] = None,
    ) -> CustomModelLLMValidation:
        """
        Start the validation of the deployment that will serve as an LLM.

        Parameters
        ----------
        deployment_id : Union[Deployment, str]
            The deployment to validate, either `Deployment` or the deployment ID.
        model : Optional[Union[Model, str]], optional
            The specific model within the deployment, either `Model` or the model ID.
            If not specified, the underlying model ID will be automatically derived from the
            deployment information.
        use_case : Optional[Union[UseCase, str]], optional
            The Use Case to link the validation to, either `UseCase` or the Use Case ID.
        name : Optional[str], optional
            The name of the validation.
        wait_for_completion : bool
            If set to `True`, the code will wait for the validation job to complete before returning
            results. If the job does not finish in 10 minutes, this method call raises a timeout
            error.
            If set to `False`, the code does not wait for the job to complete. Instead,
            `CustomModelLLMValidation.get` can be used to poll for the status of the job using
            the validation ID returned by the method.
        prediction_timeout : Optional[int], optional
            The timeout, in seconds, for the prediction API used in this custom model validation.
        prompt_column_name : Optional[str], optional
            The name of the column the deployed model uses for prompt text input.
            This value is used to call the Prediction API of the deployment.
            For LLM deployments that support the chat completion API, it is recommended to
            specify `chatModelId` instead.
        target_column_name : Optional[str], optional
            The name of the column the deployed model uses for prediction output.
            This value is used to call the Prediction API of the deployment.
            For LLM deployments that support the chat completion API, it is recommended to
            specify `chatModelId` instead.
        chat_model_id : Optional[str], optional
            The model ID to specify when calling the chat completion API of the deployment.
            If this parameter is specified, the deployment must support the chat completion API.

        Returns
        -------
        CustomModelLLMValidation
        """

        payload = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "chat_model_id": chat_model_id,
            "deployment_id": get_entity_id(deployment_id),
            "model_id": get_entity_id(model) if model else None,
            "use_case_id": get_use_case_id(use_case, is_required=False),
            "name": name,
        }
        if prediction_timeout is not None:
            payload["prediction_timeout"] = prediction_timeout  # type: ignore[assignment]
        url = f"{cls._client.domain}/{cls._path}/"
        response = cls._client.post(url, data=payload)
        if wait_for_completion:
            location = wait_for_async_resolution(cls._client, response.headers["Location"])
            return cls.from_location(location)
        return cls.from_server_data(response.json())

    def update(
        self,
        name: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        chat_model_id: Optional[str] = None,
        deployment: Optional[Union[Deployment, str]] = None,
        model: Optional[Union[Model, str]] = None,
        prediction_timeout: Optional[int] = None,
    ) -> CustomModelLLMValidation:
        """
        Update a custom model validation.

        Parameters
        ----------
        name : Optional[str], optional
            The new name of the custom model validation.
        prompt_column_name : Optional[str], optional
            The new name of the prompt column.
        target_column_name : Optional[str], optional
            The new name of the target column.
        chat_model_id : Optional[str], optional
            The new model ID to specify when calling the chat completion API of the deployment.
        deployment : Optional[Union[Deployment, str]]
            The new deployment to validate.
        model : Optional[Union[Model, str]], optional
            The new model within the deployment to validate.
        prediction_timeout : Optional[int], optional
            The new timeout, in seconds, for the prediction API used in this custom model validation.

        Returns
        -------
        CustomModelLLMValidation
        """
        payload = {
            "name": name,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "chat_model_id": chat_model_id,
            "deployment_id": get_entity_id(deployment) if deployment else None,
            "model_id": get_entity_id(model) if model else None,
            "prediction_timeout": prediction_timeout,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        response = self._client.patch(url, data=payload)
        return self.from_server_data(response.json())



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from contextlib import contextmanager
import contextvars
import logging
import re
from typing import Optional, Tuple, TYPE_CHECKING, Union
import warnings

from urllib3 import Retry

from ._version import __expected_server_version__, __version__
from .config import create_drconfig
from .context import Context, DefaultUseCase
from .errors import ClientError
from .rest import RESTClientObject

if TYPE_CHECKING:
    from requests import Response

logger = logging.getLogger(__package__)
VERSION_RE = re.compile(r"^(\d+)\.(\d+).*")

__all__ = ("Client", "get_client", "set_client", "client_configuration")

_global_client: Optional[RESTClientObject] = None
_context_client: contextvars.ContextVar[RESTClientObject] = contextvars.ContextVar("dr_rest_client")


def _create_client(
    *args, **kwargs
) -> Tuple[RESTClientObject, Optional[str], Optional[bool], Optional[str]]:
    """Return a new `RESTClientObject` and default_use_case id with optional configuration.

    Parameters
    ----------
        args : Parameters passed to ``datarobot.client.Client()``
        kwargs : Keyword arguments passed to ``datarobot.client.Client()``

    Returns
    -------
    client : RESTClientObject
        The client instance.
    default_use_case : Optional[str]
        The entity ID of the default Use Case to use with any requests made by the client.
    enable_api_consumer_tracking : Optional[bool]
        Whether to enable user metrics tracking within the datarobot module.
    trace_context : Optional[str]
        An ID or other string for identifying which code template or AI Accelerator was used to make
        a request.
    """
    drconfig = create_drconfig(*args, **kwargs)

    client = RESTClientObject.from_config(drconfig)
    if not _is_compatible_client(client):
        raise ValueError("The client is not compatible with the server version")

    return (
        client,
        drconfig.default_use_case,
        drconfig.enable_api_consumer_tracking,
        drconfig.trace_context,
    )


def _is_compatible_client(client: RESTClientObject) -> bool:
    """
    Check that this client version is not ahead of the DataRobot version that
    the server points to. There will be unsupported functionality

    Parameters
    ----------
    client : RESTClientObject

    Returns
    -------
    bool : True if client is compatible with server, False otherwise.

    Raises
    ------
    ClientError
        If the client is unable to authenticate with the DataRobot API.

    """
    try:
        server_response = client.get("version/")
    except ClientError as cerr:
        # TODO Prefer to do further error handling via BaseException.add_note but that
        # is only available in Python 3.11+
        if cerr.status_code == 401:
            w_msg_tmpl = (
                'Unable to authenticate to the server - are you sure the provided token of "{}" '
                'and endpoint of "{}" are correct? '.format(client.token, client.endpoint)
            )
        else:
            w_msg_tmpl = (
                "Error retrieving a version string from the server. "
                "Server did not reply with an API version. This may indicate the "
                "endpoint parameter `{}` is incorrect, or that the server there is "
                "incompatible with this version of the DataRobot client package. "
            )
        w_msg_tmpl += (
            "Note that if you access the DataRobot webapp at "
            "`https://app.datarobot.com`, then the correct endpoint to specify would "
            "be `https://app.datarobot.com/api/v2`."
        )
        raise ClientError(
            w_msg_tmpl.format(client.endpoint), status_code=cerr.status_code
        ) from cerr

    if not _is_compatible_version(server_response):
        return False

    _ensure_protocol_match(client, server_response)
    return True


# TODO: consider replacing this function with `packaging` library once we'll need to have a more
# complicated version checking
def _get_major_minor_version(version_string: str) -> Tuple[int, int]:
    """
    Parse DataRobot version string and return major and minor numbers.

    Parameters
    ----------
    version_string : str
        The version string returned by client or server.

    Returns
    -------
    Tuple[int, int]
    """
    major, minor = VERSION_RE.match(version_string).groups()
    return int(major), int(minor)


def _is_compatible_version(version_response: Response) -> bool:
    """
    Ensure that server and client are compatible.

    Parameters
    ----------
    version_response : request.Response
        client.get('version/') response object

    Returns
    -------
    bool : True if client and server versions are compatible, False otherwise
    """
    response_json = version_response.json()
    server_version_string = response_json["versionString"]
    if server_version_string is None:
        warn_msg = (
            "Server did not respond with a version string, you may have incompatibilities. "
            "Please check that your versions of the DataRobot application and this package "
            "are compatible"
        )
        warnings.warn(warn_msg)
        return True
    expected_server_version_string = _get_expected_server_version()
    server_major, server_minor = _get_major_minor_version(server_version_string)
    expected_major, expected_minor = _get_major_minor_version(expected_server_version_string)
    if server_major > expected_major:
        err_msg = (
            f"Client and server versions incompatible. Expected server version: {expected_server_version_string} - "
            f"Actual server version: {server_version_string}"
        )
        warnings.warn(err_msg)
        return False
    if server_major < expected_major or server_minor < expected_minor:
        warn_msg = (
            "Client version is ahead of server version, you may have incompatibilities. "
            f"Expected server version: {expected_server_version_string} - "
            f"Actual server version: {server_version_string}"
        )
        warnings.warn(warn_msg)
    return True


def _ensure_protocol_match(client: RESTClientObject, server_response: Response) -> None:
    """
    Check if server responded using the same protocol as the client endpoint configuration.
    If protocol mismatch detected - the client endpoint will be updated to https version.

    Parameters
    ----------
    client : RESTClientObject
        datarobot client instance
    server_response : request.Response
        response from 'version/' endpoint.
    """
    # Do not proceed if there was no redirect
    if not server_response.history:
        return
    # check the redirect location, if it is the https version - update client endpoint.
    location = server_response.history[0].headers["Location"]
    expected_location = client._join_endpoint("version/").replace("http://", "https://")
    if location == expected_location:
        warn_msg = (
            "Client endpoint is configured for HTTP protocol; "
            "however the server users HTTPS. HTTPS will be used."
        )
        warnings.warn(warn_msg)
        if not client.endpoint:
            raise ValueError("Client endpoint is not set and is required.")
        client.endpoint = client.endpoint.replace("http://", "https://")


def _get_expected_server_version() -> str:
    return __expected_server_version__


def _get_client_version() -> str:
    return __version__


# Create Client section
def get_client() -> RESTClientObject:
    """
    Returns the global HTTP client for the Python SDK, instantiating it
    if necessary.
    """
    return _context_client.get(_global_client) or Client()


class staticproperty(property):
    def __get__(self, instance, owner):
        return self.fget()


def set_client(client: RESTClientObject) -> Optional[RESTClientObject]:
    """
    Configure the global HTTP client for the Python SDK.
    Returns previous instance.
    """
    global _global_client  # pylint: disable=W0603
    previous = _context_client.get(_global_client)
    _context_client.set(client)
    # addressing issue for ipykernel < 6.0
    # https://github.com/ipython/ipython/issues/11565
    # fall back to _global_client
    _global_client = client
    return previous


def Client(
    token: Optional[str] = None,
    endpoint: Optional[str] = None,
    config_path: Optional[str] = None,
    connect_timeout: Optional[int] = None,
    ssl_verify: Optional[bool] = None,
    max_retries: Optional[Union[int, Retry]] = None,
    use_tcp_keepalive: Optional[bool] = None,
    token_type: Optional[str] = None,
    default_use_case: Optional[str] = None,
    enable_api_consumer_tracking: Optional[bool] = None,
    trace_context: Optional[str] = None,
) -> RESTClientObject:
    """
    Configures the global API client for the Python SDK. The client will be configured in one of
    the following ways, in order of priority.

    Notes
    -----
    Token and endpoint must be specified from one source only. This is a restriction
    to prevent token leakage if environment variables or config file are used.

    The DataRobotClientConfig params will be looking up to find the configuration parameters
    in one of the following ways,

      1. From call kwargs if specified;
      2. From a YAML file at the path specified in the ``config_path`` kwarg;
      3. From a YAML file at the path specified in the environment variables ``DATAROBOT_CONFIG_FILE``;
      4. From environment variables;
      5. From the default values in the default YAML file
         at the path `$HOME/.config/datarobot/drconfig.yaml`.

    This can also have the side effect of setting a default Use Case for client API requests.

    Parameters
    ----------
    token : Optional[str]
        API token.
    endpoint : Optional[str]
        Base URL of API.
    config_path : Optional[str]
        An alternate location of the config file.
    connect_timeout : Optional[int]
        How long the client should be willing to wait before giving up on establishing
        a connection with the server.
    ssl_verify : bool or Optional[str]
        Whether to check SSL certificate.
        Could be set to path with certificates of trusted certification authorities. Default: True.
    max_retries : int or urllib3.util.retry.Retry, optional
        Either an integer number of times to retry connection errors,
        or a `urllib3.util.retry.Retry` object to configure retries.
    use_tcp_keepalive : Optional[bool]
        Enable TCP keepalive for the client. If None is set to True. Default: None.
    token_type: Optional[str]
        Authentication token type: Token, Bearer.
        "Bearer" is for DataRobot OAuth2 token, "Token" for token generated in Developer Tools.
        Default: "Token".
    default_use_case: Optional[str]
        The entity ID of the default Use Case to use with any requests made by the client.
    enable_api_consumer_tracking: Optional[bool]
        Enable and disable user metrics tracking within the datarobot module. Default: False.
    trace_context: Optional[str]
        An ID or other string for identifying which code template or AI Accelerator was used to make
        a request.

    Returns
    -------
        The ``RESTClientObject`` instance created.
    """

    new_client, default_use_case, enable_api_consumer_tracking, trace_context = _create_client(
        token,
        endpoint,
        config_path,
        connect_timeout,
        ssl_verify,
        max_retries,
        token_type,
        default_use_case,
        enable_api_consumer_tracking,
        trace_context,
        use_tcp_keepalive,
    )

    set_client(new_client)
    Context.use_case = default_use_case
    Context.trace_context = trace_context
    Context.enable_api_consumer_tracking = enable_api_consumer_tracking

    return new_client


@contextmanager
def client_configuration(*args, **kwargs):
    """
    This context manager can be used to temporarily change the global HTTP client.

    In multithreaded scenarios, it is highly recommended to use a fresh manager object
    per thread.

    DataRobot does not recommend nesting these contexts.

    Parameters
    ----------
        args : Parameters passed to ``datarobot.client.Client()``
        kwargs : Keyword arguments passed to ``datarobot.client.Client()``

    Examples
    --------

    .. code-block:: python

        from datarobot.client import client_configuration
        from datarobot.models import Project

        with client_configuration(default_use_case=[]):
            # Interact with all accessible projects, not just those associated
            # with the current use case.
            Project.list()

        with client_configuration(token="api-key-here", endpoint="https://host-name.com"):
            # Interact with projects on a different DataRobot instance.
            Project.list()

    .. code-block:: python

        from datarobot.client import Client, client_configuration
        from datarobot.models import Project

        Client()  # Interact with DataRobot using the default configuration.
        Project.list()

        with client_configuration(config_path="/path/to/a/drconfig.yaml"):
            # Interact with DataRobot using a different configuration.
            Project.list()
    """

    contextvars_token: contextvars.Token[RESTClientObject] = None
    previous_use_case: DefaultUseCase = None
    previous_enable_api_consumer_tracking: Optional[bool] = None
    previous_trace_context: Optional[str] = None
    try:
        client, default_use_case, enable_api_consumer_tracking, trace_context = _create_client(
            *args, **kwargs
        )
        contextvars_token = _context_client.set(client)
        previous_use_case = Context.get_use_case(raw=True)
        previous_enable_api_consumer_tracking = Context.enable_api_consumer_tracking
        previous_trace_context = Context.trace_context
        Context.use_case = default_use_case
        Context.enable_api_consumer_tracking = enable_api_consumer_tracking
        Context.trace_context = trace_context
        yield
    finally:
        if contextvars_token is not None:
            _context_client.reset(contextvars_token)
        Context.use_case = previous_use_case
        Context.enable_api_consumer_tracking = previous_enable_api_consumer_tracking
        Context.trace_context = previous_trace_context



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import NotRequired, TypedDict

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.models.visualai.images import Image
from datarobot.utils import from_api
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

__all__ = ["ImageAugmentationOptions", "ImageAugmentationList", "ImageAugmentationSample"]


class TransformationParam(TypedDict):
    name: str
    min_value: NotRequired[Union[int, float]]
    max_value: NotRequired[Union[int, float]]
    current_value: NotRequired[Union[int, float]]


class Transformation(TypedDict):
    name: str
    enabled: NotRequired[bool]
    params: NotRequired[List[TransformationParam]]


Transformations = List[Transformation]


class ImageAugmentationOptions(APIObject):
    """A List of all supported Image Augmentation Transformations for a project.
    Includes additional information about minimum, maximum, and default values
    for a transformation.

    Attributes
    ----------
    name: str
        The name of the augmentation list
    project_id: str
        The project containing the image data to be augmented
    min_transformation_probability: float
        The minimum allowed value for transformation probability.
    current_transformation_probability: float
        Default setting for probability that each transformation will be applied to an image.
    max_transformation_probability: float
        The maximum allowed value for transformation probability.
    min_number_of_new_images: int
         The minimum allowed number of new rows to add for each existing row
    current_number_of_new_images: int
         The default number of new rows to add for each existing row
    max_number_of_new_images: int
         The maximum allowed number of new rows to add for each existing row
    transformations: list[dict]
        List of transformations to possibly apply to each image
    """

    _get_path = "imageAugmentationOptions/{pid}"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("project_id"): String(),
            t.Key("min_transformation_probability"): t.Float(),
            t.Key("max_transformation_probability"): t.Float(),
            t.Key("current_transformation_probability"): t.Float(),
            t.Key("min_number_of_new_images"): Int(),
            t.Key("current_number_of_new_images"): Int(),
            t.Key("max_number_of_new_images"): Int(),
            t.Key("transformations", optional=True): t.List(t.Dict().allow_extra("*")),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        project_id: str,
        min_transformation_probability: float,
        current_transformation_probability: float,
        max_transformation_probability: float,
        min_number_of_new_images: int,
        current_number_of_new_images: int,
        max_number_of_new_images: int,
        transformations: Optional[Transformations] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.project_id = project_id
        self.min_transformation_probability = min_transformation_probability
        self.current_transformation_probability = current_transformation_probability
        self.max_transformation_probability = max_transformation_probability
        self.min_number_of_new_images = min_number_of_new_images
        self.current_number_of_new_images = current_number_of_new_images
        self.max_number_of_new_images = max_number_of_new_images
        self.transformations = transformations

    def __repr__(self) -> str:
        return (
            "datarobot.models.visualai.ImageAugmentationOptions("
            "id={0.id}, "
            "name={0.name}, "
            "project_id={0.project_id}, "
            "min_transformation_probability={0.min_transformation_probability}, "
            "max_transformation_probability={0.max_transformation_probability}, "
            "current_transformation_probability={0.current_transformation_probability}, "
            "min_number_of_new_images={0.min_number_of_new_images})"
            "max_number_of_new_images={0.max_number_of_new_images})"
            "current_number_of_new_images={0.current_number_of_new_images})"
        ).format(self)

    @classmethod
    def get(cls, project_id: str) -> "ImageAugmentationOptions":
        """
        Returns a list of all supported transformations for the given
        project

        :param project_id: sting
            The id of the project for which to return the list of supported transformations.

        :return:
          ImageAugmentationOptions
           A list containing all the supported transformations for the project.
        """
        path = cls._get_path.format(pid=project_id)
        server_data = cls._client.get(path)
        return cls.from_server_data(server_data.json())


class ImageAugmentationList(APIObject):
    """A List of Image Augmentation Transformations

    Attributes
    ----------
    name: str
        The name of the augmentation list
    project_id: str
        The project containing the image data to be augmented
    feature_name: Optional[str]
        name of the feature that the augmentation list is associated with
    in_use: bool
        Whether this is the list that will passed in to every blueprint during blueprint generation
        before autopilot
    initial_list: bool
        True if this is the list to be used during training to produce augmentations
    transformation_probability: float
        Probability that each transformation will be applied to an image.  Value should be
        between 0.01 - 1.0.
    number_of_new_images: int
         Number of new rows to add for each existing row
    transformations: List[Dict]
        List of transformations to possibly apply to each image
    samples_id: str
        Id of last image augmentation sample generated for image augmentation list.
    """

    _list_path = "imageAugmentationLists/"
    _referenced_by_id_path = "imageAugmentationLists/{id}/"
    _create_path = "imageAugmentationLists/"
    _samples_path = "imageAugmentationLists/{id}/samples/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("project_id"): String(),
            t.Key("feature_name", optional=True): String(),
            t.Key("in_use", optional=True): t.Bool(),
            t.Key("initial_list", optional=True): t.Bool(),
            t.Key("transformation_probability", optional=True): t.Float(),
            t.Key("number_of_new_images", optional=True, default=1): Int(),
            t.Key("transformations", optional=True): t.List(t.Dict().allow_extra("*")),
            t.Key("samples_id", optional=True): String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        project_id: str,
        feature_name: Optional[str] = None,
        in_use: bool = False,
        initial_list: bool = False,
        transformation_probability: float = 0.0,
        number_of_new_images: int = 1,
        transformations: Optional[Transformations] = None,
        samples_id: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.project_id = project_id
        self.feature_name = feature_name
        self.in_use = in_use
        self.initial_list = initial_list
        self.transformation_probability = transformation_probability
        self.number_of_new_images = number_of_new_images
        self.transformations = transformations
        self.samples_id = samples_id

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ImageAugmentationList)
            and self.id == other.id
            and self.name == other.name
            and self.project_id == other.project_id
            and self.feature_name == other.feature_name
            and self.in_use == other.in_use
            and self.initial_list == other.initial_list
            and self.transformation_probability == other.transformation_probability
            and self.number_of_new_images == other.number_of_new_images
            and self.transformations == other.transformations
            and self.samples_id == other.samples_id
        )

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __repr__(self) -> str:
        return (
            "datarobot.models.visualai.ImageAugmentationList("
            "id={0.id}, "
            "name='{0.name}', "
            "project_id={0.project_id}, "
            "feature_name='{0.feature_name}', "
            "in_use={0.in_use}, "
            "initial_list={0.initial_list}, "
            "transformation_probability={0.transformation_probability}, "
            "number_of_new_images={0.number_of_new_images}, "
            "samples_id={0.samples_id})"
        ).format(self)

    @classmethod
    def create(
        cls,
        name: str,
        project_id: str,
        feature_name: Optional[str] = None,
        initial_list: bool = False,
        transformation_probability: float = 0.0,
        number_of_new_images: int = 1,
        transformations: Optional[Transformations] = None,
        samples_id: Optional[str] = None,
    ) -> "ImageAugmentationList":
        """
        create a new image augmentation list
        """
        data = {
            "name": name,
            "project_id": project_id,
            "feature_name": feature_name,
            "initial_list": initial_list,
            "samples_id": samples_id,
            "transformation_probability": transformation_probability,
            "number_of_new_images": number_of_new_images,
            "transformations": transformations,
        }
        server_data = cls._client.post(cls._create_path, data=data)
        list_id = server_data.json()["augmentationListId"]
        return cls.get(list_id)

    @classmethod
    def get(cls, list_id: str) -> "ImageAugmentationList":
        path = cls._referenced_by_id_path.format(id=list_id)
        server_data = cls._client.get(path)
        return cls.from_server_data(server_data.json())

    @classmethod
    def delete(cls, list_id: str) -> None:
        path = cls._referenced_by_id_path.format(id=list_id)
        cls._client.delete(path)

    @classmethod
    def list(
        cls, project_id: str, feature_name: Optional[str] = None
    ) -> List["ImageAugmentationList"]:
        """
        List Image Augmentation Lists present in a project.

        Parameters
        ----------
        project_id : str
            Project Id to retrieve augmentation lists for.
        feature_name : Optional[str]
            If passed, the response will only include Image Augmentation Lists active for the
            provided feature name.

        Returns
        -------
        list[ImageAugmentationList]
        """
        parameters = {"project_id": project_id}
        if feature_name is not None:
            parameters["feature_name"] = feature_name
        data = unpaginate(cls._list_path, parameters, cls._client)
        return [cls.from_server_data(item) for item in data]

    def update(
        self,
        name: Optional[str] = None,
        feature_name: Optional[str] = None,
        initial_list: Optional[bool] = None,
        transformation_probability: Optional[float] = None,
        number_of_new_images: Optional[int] = None,
        transformations: Optional[Transformations] = None,
    ) -> "ImageAugmentationList":
        """
        Update one or multiple attributes of the Image Augmentation List in the DataRobot backend
        as well on this object.

        Parameters
        ----------
        name : Optional[str]
            New name of the feature list.
        feature_name : Optional[str]
            The new feature name for which the Image Augmentation List is effective.
        initial_list : Optional[bool]
            New flag that indicates whether this list will be used during Autopilot to perform
            image augmentation.
        transformation_probability : Optional[float]
            New probability that each enabled transformation will be applied to an image.
            This does not apply to Horizontal or Vertical Flip, which are always set to 50%.
        number_of_new_images : Optional[int]
            New number of new rows to add for each existing row, updating the existing augmentation
            list.
        transformations : Optional[list]
            New list of Transformations to possibly apply to each image.

        Returns
        -------
        ImageAugmentationList
            Reference to self. The passed values will be updated in place.
        """
        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if feature_name is not None:
            params["feature_name"] = feature_name
        if initial_list is not None:
            params["initial_list"] = initial_list
        if transformation_probability is not None:
            params["transformation_probability"] = transformation_probability
        if number_of_new_images is not None:
            params["number_of_new_images"] = number_of_new_images
        if transformations is not None:
            params["transformations"] = transformations
        self._client.patch(self._create_path + str(self.id) + "/", data=params)

        for param, value in params.items():
            case_converted = from_api(value)
            setattr(self, param, case_converted)
        return self

    def retrieve_samples(self) -> List["ImageAugmentationSample"]:
        """
        Lists already computed image augmentation sample for image augmentation list.
        Returns samples only if they have been already computed. It does not initialize computation.

        Returns
        -------
        List of class ImageAugmentationSample
        """
        return ImageAugmentationSample.list(auglist_id=self.id)

    def compute_samples(self, max_wait: int = 600) -> List["ImageAugmentationSample"]:
        """
        Initializes computation and retrieves list of image augmentation samples
        for image augmentation list. If samples exited prior to this call method,
        this will compute fresh samples and return latest version of samples.

        Returns
        -------
        List of class ImageAugmentationSample
        """
        url = self._samples_path.format(id=self.id)
        response = self._client.post(url, data={"numberOfRows": self.number_of_new_images})
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait=max_wait)
        return self.retrieve_samples()


class ImageAugmentationSample(APIObject):
    """
    A preview of the type of images that augmentations will create during training.

    Attributes
    ----------
    sample_id : ObjectId
        The id of the augmentation sample, used to group related images together
    image_id : ObjectId
        A reference to the Image which can be used to retrieve the image binary
    project_id : ObjectId
        A reference to the project containing the image
    original_image_id : ObjectId
        A reference to the original image that generated this image in the case of an augmented
        image.  If this is None it signifies this is an original image
    height : int
        Image height in pixels
    width : int
        Image width in pixels
    """

    _compute_path = "imageAugmentationSamples/"
    _list_by_samples_path = "imageAugmentationSamples/{sample_id}/"
    _list_by_auglist_path = "imageAugmentationLists/{auglist_id}/samples/"
    _converter = t.Dict(
        {
            t.Key("image_id"): String(),
            t.Key("project_id"): String(),
            t.Key("height"): Int(),
            t.Key("width"): Int(),
            t.Key("original_image_id", optional=True): String(),
            t.Key("sample_id", optional=True): String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        image_id: str,
        project_id: str,
        height: int,
        width: int,
        original_image_id: Optional[str] = None,
        sample_id: Optional[str] = None,
    ) -> None:
        self.sample_id = sample_id
        self.image_id = image_id
        self.project_id = project_id
        self.original_image_id = original_image_id
        self.height = height
        self.width = width
        self.image = Image(image_id=image_id, project_id=project_id, height=height, width=width)

    def __repr__(self) -> str:
        return (
            "datarobot.models.visualai.ImageAugmentationSample("
            "image_id={0.image_id}, "
            "project_id={0.project_id}, "
            "height={0.height}, "
            "width={0.width}, "
            "original_image_id={0.original_image_id})"
            "sample_id={0.sample_id}, "
        ).format(self)

    @classmethod
    def list(cls, auglist_id: Optional[str] = None) -> List["ImageAugmentationSample"]:
        """Return a list of ImageAugmentationSample objects.

        Parameters
        ----------
        auglist_id: str
            ID for augmentation list to retrieve samples for

        Returns
        -------
        List of class ImageAugmentationSample
        """
        path = cls._list_by_auglist_path.format(auglist_id=auglist_id)

        result = cls._client.get(path)
        r_data = result.json()
        ret = []
        for sample_data in r_data.get("data", []):
            sample = cls.from_server_data(sample_data)
            ret.append(sample)
        return ret



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# pylint: disable=too-many-lines
from __future__ import annotations

import csv
import datetime
import io
from typing import cast, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

import pandas as pd
import trafaret as t

from datarobot import errors
from datarobot._compat import Int, String
from datarobot.models.batch_job import (
    AbstractBatchJob,
    CsvSettings,
    IntakeSettings,
    OutputSettings,
    recognize_sourcedata,
    Schedule,
)
from datarobot.models.credential import Credential
from datarobot.models.dataset import Dataset
from datarobot.models.prediction_explanations import PredictionExplanationsMode
from datarobot.models.project import Project
from datarobot.utils import (
    deprecation_warning,
    get_id_from_response,
    logger,
    pagination,
    parse_time,
    to_api,
)

from ..enums import DEFAULT_TIMEOUT, IntakeAdapters, JOB_TYPE, OutputAdapters, QUEUE_STATUS
from .api_object import APIObject

LOG = logger.get_logger(__name__)

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.models.deployment import Deployment
    from datarobot.models.model import DatetimeModel, Model

    DeploymentType = Union[str, Deployment]
    ModelType = Union[str, Model, DatetimeModel]

    class TimeSeriesSettings(TypedDict, total=False):
        type: str
        forecast_point: Optional[datetime.datetime]
        predictions_start_date: Optional[datetime.datetime]
        predictions_end_date: Optional[datetime.datetime]
        relax_known_in_advance_features_check: bool

    class PredictionInstance(TypedDict):
        hostName: str
        sslEnabled: Optional[bool]
        datarobotKey: Optional[str]
        apiKey: Optional[str]


class MissingType:
    pass


MISSING = MissingType()


class BatchPredictionJob(AbstractBatchJob):
    """
    A Batch Prediction Job is used to score large data sets on
    prediction servers using the Batch Prediction API.

    Attributes
    ----------
    id : str
        the id of the job
    """

    _job_spec = t.Dict(
        {
            t.Key("num_concurrent"): Int(),
            t.Key("threshold_high", optional=True): t.Float(),
            t.Key("threshold_low", optional=True): t.Float(),
            t.Key("explanation_class_names", optional=True): t.List(t.String),
            t.Key("explanation_num_top_classes", optional=True): t.Int(),
            t.Key("deployment_id", optional=True): String(),
            t.Key("model_id", optional=True): String(),
            t.Key("passthrough_columns", optional=True): t.List(String(allow_blank=True)),
            t.Key("passthrough_columns_set", optional=True): String(),
            t.Key("max_explanations", optional=True): Int(),
            t.Key("max_ngram_explanations", optional=True): t.Int(gte=0) | t.Atom("all"),
            t.Key("explanation_algorithm", optional=True): String(),
            t.Key("prediction_threshold", optional=True): t.Float(gte=0.0, lte=1.0) | t.Null(),
            t.Key("prediction_warning_enabled", optional=True): t.Bool(),
            t.Key("intake_settings", optional=True): t.Dict().allow_extra("*"),
            t.Key("output_settings", optional=True): t.Dict().allow_extra("*"),
            t.Key("timeseries_settings", optional=True): t.Dict().allow_extra("*"),
        }
    ).allow_extra("*")
    _links = t.Dict(
        {t.Key("download", optional=True): String(allow_blank=True), t.Key("self"): String()}
    ).allow_extra("*")
    _converter_extra = t.Dict(
        {
            t.Key("percentage_completed"): t.Float(),
            t.Key("elapsed_time_sec"): Int(),
            t.Key("links"): _links,
            t.Key("job_spec"): _job_spec,
            t.Key("status_details"): String(),
        }
    ).allow_extra("*")

    _converter_common = t.Dict(
        {
            t.Key("id", optional=True): String,
            t.Key("status", optional=True): t.Enum(
                QUEUE_STATUS.ABORTED,
                QUEUE_STATUS.COMPLETED,
                QUEUE_STATUS.RUNNING,
                QUEUE_STATUS.INITIALIZING,
                "FAILED",
            ),
            t.Key("project_id", optional=True): String,
            t.Key("is_blocked", optional=True): t.Bool,
        }
    )

    _timeseries_settings = (
        t.Dict(
            {
                t.Key("type"): t.Atom("forecast"),
                t.Key("forecast_point", optional=True): parse_time,
                t.Key("relax_known_in_advance_features_check", optional=True): t.Bool(),
            }
        )
        | t.Dict(
            {
                t.Key("type"): t.Atom("historical"),
                t.Key("predictions_start_date", optional=True): parse_time,
                t.Key("predictions_end_date", optional=True): parse_time,
                t.Key("relax_known_in_advance_features_check", optional=True): t.Bool(),
            }
        )
        | t.Dict(
            {
                t.Key("type"): t.Atom("training"),
                t.Key("relax_known_in_advance_features_check", optional=True): t.Bool(),
            }
        )
    )

    _prediction_instance = t.Dict(
        {
            t.Key("hostName"): String(),
            t.Key("sslEnabled", optional=True): t.Bool(),
            t.Key("datarobotKey", optional=True): String(),
            t.Key("apiKey", optional=True): String(),
        }
    )

    _column_names_remapping = t.Mapping(String, t.Or(String, t.Null()))

    @classmethod
    def _job_type(cls) -> str:
        return cast(str, JOB_TYPE.BATCH_PREDICTIONS)

    @classmethod
    def _jobs_path(cls) -> str:
        return "batchPredictions/"

    @classmethod
    def _job_path(
        cls,
        project_id: str,
        job_id: str,
    ) -> str:
        return f"batchPredictions/{job_id}/"

    @classmethod
    def _from_existing_path(cls) -> str:
        return "batchPredictions/fromExisting/"

    @classmethod
    def _score(
        cls,
        deployment_id: Optional[str] = None,
        model_id: Optional[str] = None,
        intake_settings: Optional[IntakeSettings] = None,
        output_settings: Optional[OutputSettings] = None,
        csv_settings: Optional[CsvSettings] = None,
        timeseries_settings: Optional[TimeSeriesSettings] = None,
        num_concurrent: Optional[int] = None,
        chunk_size: Optional[Union[int, str]] = None,
        passthrough_columns: Optional[List[str]] = None,
        passthrough_columns_set: Optional[str] = None,
        max_explanations: Optional[int] = None,
        max_ngram_explanations: Optional[Union[int, str]] = None,
        explanation_algorithm: Optional[str] = None,
        threshold_high: Optional[float] = None,
        threshold_low: Optional[float] = None,
        prediction_threshold: Optional[float] = None,
        prediction_warning_enabled: Optional[bool] = None,
        include_prediction_status: bool = False,
        skip_drift_tracking: bool = False,
        prediction_instance: Optional[PredictionInstance] = None,
        abort_on_error: bool = True,
        column_names_remapping: Optional[Dict[str, str]] = None,
        include_probabilities: bool = True,
        include_probabilities_classes: Optional[List[str]] = None,
        download_timeout: Optional[int] = 120,
        download_read_timeout: Optional[int] = 660,
        upload_read_timeout: Optional[int] = DEFAULT_TIMEOUT.UPLOAD,
        explanations_mode: Optional[PredictionExplanationsMode] = None,
    ) -> BatchPredictionJob:
        """Common logic for leaderboard and deployment scoring."""

        if deployment_id:
            # TODO - Make this it's own TypedDict class
            job_data: Dict[str, Union[str, int, float, bool]] = {"deploymentId": deployment_id}
        else:
            job_data: Dict[str, Union[str, int, float, bool]] = {"modelId": model_id}

        if num_concurrent is not None:
            job_data["numConcurrent"] = int(num_concurrent)

        if chunk_size is not None:
            job_data["chunkSize"] = chunk_size

        if max_explanations is not None:
            job_data["maxExplanations"] = int(max_explanations)

        if max_ngram_explanations is not None:
            job_data["maxNgramExplanations"] = max_ngram_explanations

        if explanation_algorithm is not None:
            job_data["explanationAlgorithm"] = explanation_algorithm

        if threshold_high is not None:
            job_data["thresholdHigh"] = float(threshold_high)

        if threshold_low is not None:
            job_data["thresholdLow"] = float(threshold_low)

        if prediction_threshold is not None:
            job_data["predictionThreshold"] = float(prediction_threshold)

        if explanations_mode is not None:
            job_data.update(explanations_mode.get_api_parameters(batch_route=True))

        if include_prediction_status:
            job_data["includePredictionStatus"] = bool(include_prediction_status)

        if skip_drift_tracking:
            job_data["skipDriftTracking"] = bool(skip_drift_tracking)

        if not abort_on_error:
            job_data["abortOnError"] = bool(abort_on_error)

        if not include_probabilities:
            job_data["includeProbabilities"] = bool(include_probabilities)

        if prediction_warning_enabled is not None:
            job_data["predictionWarningEnabled"] = bool(prediction_warning_enabled)

        if passthrough_columns_set == "all":
            job_data["passthroughColumnsSet"] = "all"
        elif passthrough_columns is not None:
            job_data["passthroughColumns"] = passthrough_columns

        if prediction_instance is not None:
            cls._prediction_instance.check(prediction_instance)
            job_data["predictionInstance"] = to_api(prediction_instance)

        if column_names_remapping is not None:
            cls._column_names_remapping.check(column_names_remapping)
            job_data["columnNamesRemapping"] = dict(column_names_remapping)

        if include_probabilities_classes:
            job_data["includeProbabilitiesClasses"] = list(include_probabilities_classes)

        if timeseries_settings is not None:
            timeseries_settings = cls._timeseries_settings.check(timeseries_settings)

            if timeseries_settings["type"] == "historical":
                start_date = timeseries_settings.get("predictions_start_date")
                end_date = timeseries_settings.get("predictions_end_date")

                if start_date and not isinstance(start_date, datetime.datetime):
                    raise ValueError(
                        "The value provided for predictions_start_date was not a valid format."
                    )

                if end_date and not isinstance(end_date, datetime.datetime):
                    raise ValueError(
                        "The value provided for predictions_end_date was not a valid format."
                    )

                if start_date and not end_date:
                    raise ValueError(
                        "You must also provide predictions_end_date if you "
                        "provide predictions_start_date"
                    )

                if start_date and end_date and start_date > end_date:
                    raise ValueError(
                        "The value provided for predictions_end_date must be after "
                        "the value provided for predictions_start_date"
                    )

            if timeseries_settings["type"] == "forecast":
                forecast_point = timeseries_settings.get("forecast_point")

                if forecast_point and not isinstance(forecast_point, datetime.datetime):
                    raise ValueError(
                        "The value provided for forecast_point was not a valid format."
                    )

            job_data["timeseriesSettings"] = to_api(timeseries_settings)

        # validate input settings, return a copy not original
        intake_settings = cls.validate_intake_settings(intake_settings)
        intake_file = None
        if intake_settings["type"] == IntakeAdapters.LOCAL_FILE:
            intake_file = recognize_sourcedata(intake_settings.pop("file"), "prediction.csv")
        job_data["intakeSettings"] = to_api(intake_settings)

        output_file = None

        # Validate the output settings

        output_settings = cls.validate_output_settings(output_settings)
        if output_settings["type"] == OutputAdapters.LOCAL_FILE:
            if output_settings.get("path") is not None:
                output_file = open(  # pylint: disable=consider-using-with
                    output_settings.pop("path"), "wb"
                )
        job_data["outputSettings"] = to_api(output_settings)

        if csv_settings is not None:
            cls._csv_settings.check(csv_settings)
            job_data["csvSettings"] = to_api(csv_settings)

        response = cls._client.post(url=cls._jobs_path(), json=job_data)

        job_response = response.json()
        job_id = get_id_from_response(response)

        upload_thread = None

        if intake_file is not None:
            upload_thread = cls._upload_intake_file(
                intake_file, intake_settings, job_response, upload_read_timeout, output_file
            )

        job = BatchPredictionJob.get(job_id)

        if output_file is not None:
            # We must download the result to `output_file`
            # And clean up any thread we spawned during uploading
            try:
                job.download(
                    output_file, timeout=download_timeout, read_timeout=download_read_timeout
                )
            finally:
                output_file.close()
                if upload_thread is not None:
                    upload_thread.join()

        return job

    @classmethod
    def score(
        cls,
        deployment: DeploymentType,
        intake_settings: Optional[IntakeSettings] = None,
        output_settings: Optional[OutputSettings] = None,
        csv_settings: Optional[CsvSettings] = None,
        timeseries_settings: Optional[TimeSeriesSettings] = None,
        num_concurrent: Optional[int] = None,
        chunk_size: Optional[Union[int, str]] = None,
        passthrough_columns: Optional[List[str]] = None,
        passthrough_columns_set: Optional[str] = None,
        max_explanations: Optional[int] = None,
        max_ngram_explanations: Optional[Union[int, str]] = None,
        explanation_algorithm: Optional[str] = None,
        threshold_high: Optional[float] = None,
        threshold_low: Optional[float] = None,
        prediction_threshold: Optional[float] = None,
        prediction_warning_enabled: Optional[bool] = None,
        include_prediction_status: bool = False,
        skip_drift_tracking: bool = False,
        prediction_instance: Optional[PredictionInstance] = None,
        abort_on_error: bool = True,
        column_names_remapping: Optional[Dict[str, str]] = None,
        include_probabilities: bool = True,
        include_probabilities_classes: Optional[List[str]] = None,
        download_timeout: Optional[int] = 120,
        download_read_timeout: Optional[int] = 660,
        upload_read_timeout: Optional[int] = DEFAULT_TIMEOUT.UPLOAD,
        explanations_mode: Optional[PredictionExplanationsMode] = None,
    ) -> BatchPredictionJob:
        """
        Create new batch prediction job, upload the scoring dataset and
        return a batch prediction job.

        The default intake and output options are both `localFile` which
        requires the caller to pass the `file` parameter and either
        download the results using the `download()` method afterwards or
        pass a path to a file where the scored data will be downloaded to
        afterwards.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        intake_settings : Optional[IntakeSettings]
            A dict configuring how data is coming from. Supported options:

                - type : str, either `localFile`, `s3`, `azure`, `gcp`, `dataset`, `jdbc`
                  `snowflake`, `synapse`, `bigquery`, or `datasphere`

            Note that to pass a dataset, you not only need to specify the `type` parameter
            as `dataset`, but you must also set the `dataset` parameter as a
            `dr.Dataset` object.

            To score from a local file, add the this parameter to the
            settings:

                - file : file-like object, string path to file or a
                  pandas.DataFrame of scoring data

            To score from S3, add the next parameters to the settings:

                - url : str, the URL to score (e.g.: `s3://bucket/key`)
                - credential_id : Optional[str]
                - endpoint_url : Optional[str], any non-default endpoint
                  URL for S3 access (omit to use the default)

            To score from JDBC, add the next parameters to the settings:

                - data_store_id : str, the ID of the external data store connected
                  to the JDBC data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - query : str (optional if `table`, `schema` and/or `catalog` is specified),
                  a self-supplied SELECT statement of the data set you wish to predict.
                - table : str (optional if `query` is specified),
                  the name of specified database table.
                - schema : str (optional if `query` is specified),
                  the name of specified database schema.
                - catalog : str  (optional if `query` is specified),
                  (new in v2.22) the name of specified database catalog.
                - fetch_size : Optional[int],
                  Changing the `fetchSize` can be used to balance throughput and memory
                  usage.
                - credential_id : Optional[str] the ID of the credentials holding
                  information about a user with read-access to the JDBC data source (see
                  :ref:`Credentials <credentials-api-doc>`).

            To score from Datasphere, add the next parameters to the settings:

                - `data_store_id` : str, the ID of the external data store connected to
                  the Datasphere data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - `table` : str,  the name of specified database table.
                - `schema` : str, the name of specified database schema.
                - `credential_id` : str, the ID of the credentials holding information about
                  a user with read-access to the Datasphere data source (see
                  :ref:`Credentials <credentials-api-doc>`).

        output_settings : Optional[OutputSettings]
            A dict configuring how scored data is to be saved. Supported
            options:

                - type : str, either `localFile`, `s3`, `azure`, `gcp`, `jdbc`,
                  `snowflake`, `synapse`, `bigquery`, or `datasphere`

            To save scored data to a local file, add this parameters to the
            settings:

                - path : Optional[str], path to save the scored data
                  as CSV. If a path is not specified, you must download
                  the scored data yourself with `job.download()`.
                  If a path is specified, the call will block until the
                  job is done. if there are no other jobs currently
                  processing for the targeted prediction instance,
                  uploading, scoring, downloading will happen in parallel
                  without waiting for a full job to complete. Otherwise,
                  it will still block, but start downloading the scored
                  data as soon as it starts generating data. This is the
                  fastest method to get predictions.

            To save scored data to S3, add the next parameters to the settings:

                - url : str, the URL for storing the results
                  (e.g.: `s3://bucket/key`)
                - credential_id : Optional[str]
                - endpoint_url : Optional[str], any non-default endpoint
                  URL for S3 access (omit to use the default)

            To save scored data to JDBC, add the next parameters to the settings:

                - `data_store_id` : str, the ID of the external data store connected to
                  the JDBC data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - `table` : str,  the name of specified database table.
                - `schema` : Optional[str], the name of specified database schema.
                - `catalog` : Optional[str], (new in v2.22) the name of specified database
                  catalog.
                - `statement_type` : str, the type of insertion statement to create,
                  one of ``datarobot.enums.AVAILABLE_STATEMENT_TYPES``.
                - `update_columns` : list(string) (optional),  a list of strings containing
                  those column names to be updated in case `statement_type` is set to a
                  value related to update or upsert.
                - `where_columns` : list(string) (optional), a list of strings containing
                  those column names to be selected in case `statement_type` is set to a
                  value related to insert or update.
                - `credential_id` : str, the ID of the credentials holding information about
                  a user with write-access to the JDBC data source (see
                  :ref:`Credentials <credentials-api-doc>`).

            To save scored data to Datasphere, add the following parameters to the settings:

                - `data_store_id` : str, the ID of the external data store connected to
                  the Datasphere data source (see
                  :ref:`Database Connectivity <database-connectivity-overview>`).
                - `table` : str,  the name of specified database table.
                - `schema` : str, the name of specified database schema.
                - `credential_id` : str, the ID of the credentials holding information about
                  a user with write-access to the Datasphere data source (see
                  :ref:`Credentials <credentials-api-doc>`).

        csv_settings : Optional[CsvSettings]
            CSV intake and output settings. Supported options:

            - `delimiter` : str (optional, default `,`), fields are delimited by
              this character. Use the string `tab` to denote TSV (TAB separated values).
              Must be either a one-character string or the string `tab`.
            - `quotechar` : str (optional, default `"`), fields containing the
              delimiter must be quoted using this character.
            - `encoding` : str (optional, default `utf-8`), encoding for the CSV
              files. For example (but not limited to): `shift_jis`, `latin_1` or
              `mskanji`.

        timeseries_settings : Optional[TimeSeriesSettings]
            Configuration for time-series scoring. Supported options:

            - `type` : str, must be `forecast` or `historical` (default if
              not passed is `forecast`). `forecast` mode makes predictions using
              `forecast_point` or rows in the dataset without target. `historical`
              enables bulk prediction mode which calculates predictions for all
              possible forecast points and forecast distances in the dataset within
              `predictions_start_date`/`predictions_end_date` range.
            - `forecast_point` : Optional[datetime.datetime], forecast point for the dataset,
              used for the forecast predictions, by default value will be inferred
              from the dataset. May be passed if ``timeseries_settings.type=forecast``.
            - `predictions_start_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `predictions_end_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `relax_known_in_advance_features_check` : bool, (default `False`).
              If True, missing values in the known in advance features are allowed
              in the forecast window at the prediction time. If omitted or False,
              missing values are not allowed.

        num_concurrent : Optional[int]
            Number of concurrent chunks to score simultaneously. Defaults to
            the available number of cores of the deployment. Lower it to leave
            resources for real-time scoring.

        chunk_size : str or Optional[int]
            Which strategy should be used to determine the chunk size.
            Can be either a named strategy or a fixed size in bytes.
            - auto: use fixed or dynamic based on flipper
            - fixed: use 1MB for explanations, 5MB for regular requests
            - dynamic: use dynamic chunk sizes
            - int: use this many bytes per chunk

        passthrough_columns : list[string] (optional)
            Keep these columns from the scoring dataset in the scored dataset.
            This is useful for correlating predictions with source data.

        passthrough_columns_set : Optional[str]
            To pass through every column from the scoring dataset, set this to
            `all`. Takes precedence over `passthrough_columns` if set.

        max_explanations : Optional[int]
            Compute prediction explanations for this amount of features.

        max_ngram_explanations : int or str (optional)
            Compute text explanations for this amount of ngrams. Set to `all` to return all ngram
            explanations, or set to a positive integer value to limit the amount of ngram
            explanations returned. By default no ngram explanations will be computed and returned.

        threshold_high : Optional[float]
            Only compute prediction explanations for predictions above this
            threshold. Can be combined with `threshold_low`.

        threshold_low : Optional[float]
            Only compute prediction explanations for predictions below this
            threshold. Can be combined with `threshold_high`.

        explanations_mode : PredictionExplanationsMode, optional
            Mode of prediction explanations calculation for multiclass and clustering models, if not
            specified - server default is to explain only the predicted class, identical to passing
            TopPredictionsMode(1).

        prediction_warning_enabled : Optional[bool]
            Add prediction warnings to the scored data. Currently only
            supported for regression models.

        include_prediction_status : Optional[bool]
            Include the prediction_status column in the output, defaults to `False`.

        skip_drift_tracking : Optional[bool]
            Skips drift tracking on any predictions made from this job. This is useful when running
            non-production workloads to not affect drift tracking and cause unnecessary alerts.
            Defaults to `False`.

        prediction_instance : Optional[PredictionInstance]
            Defaults to instance specified by deployment or system configuration.
            Supported options:

                - `hostName` : str
                - `sslEnabled` : boolean (optional, default `true`). Set to `false` to
                  run prediction requests from the batch prediction job without SSL.
                - `datarobotKey` : Optional[str], if running a job against a prediction
                  instance in the Managed AI Cloud, you must provide the organization level
                  DataRobot-Key
                - `apiKey` : Optional[str], by default, prediction requests will use the
                  API key of the user that created the job. This allows you to make requests
                  on behalf of other users.

        abort_on_error : Optional[bool]
             Default behavior is to abort the job if too many rows fail scoring. This will free
             up resources for other jobs that may score successfully. Set to `false` to
             unconditionally score every row no matter how many errors are encountered.
             Defaults to `True`.

        column_names_remapping : Optional[Dict[str, str]]
            Mapping with column renaming for output table. Defaults to `{}`.

        include_probabilities : Optional[bool]
            Flag that enables returning of all probability columns. Defaults to `True`.

        include_probabilities_classes : list (optional)
            List the subset of classes if a user doesn't want all the classes. Defaults to `[]`.

        download_timeout : Optional[int]
            .. versionadded:: 2.22

            If using localFile output, wait this many seconds for the download to become
            available. See `download()`.

        download_read_timeout : Optional[int], default 660
            .. versionadded:: 2.22

            If using localFile output, wait this many seconds for the server to respond
            between chunks.

        upload_read_timeout: Optional[int], default 600
            .. versionadded:: 2.28

            If using localFile intake, wait this many seconds for the server to respond
            after whole dataset upload.

        prediction_threshold: Optional[float]
            .. versionadded:: 3.4.0

            Threshold is the point that sets the class boundary for a predicted value. The model
            classifies an observation below the threshold as FALSE, and an observation above the
            threshold as TRUE. In other words, DataRobot automatically assigns the positive class
            label to any prediction exceeding the threshold.
            This value can be set between 0.0 and 1.0.

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        try:
            dep_id = cast("Deployment", deployment).id
        except AttributeError:
            dep_id = cast(str, deployment)

        return cls._score(
            deployment_id=dep_id,
            intake_settings=intake_settings,
            output_settings=output_settings,
            csv_settings=csv_settings,
            timeseries_settings=timeseries_settings,
            num_concurrent=num_concurrent,
            chunk_size=chunk_size,
            passthrough_columns=passthrough_columns,
            passthrough_columns_set=passthrough_columns_set,
            max_explanations=max_explanations,
            max_ngram_explanations=max_ngram_explanations,
            explanation_algorithm=explanation_algorithm,
            threshold_high=threshold_high,
            threshold_low=threshold_low,
            prediction_threshold=prediction_threshold,
            prediction_warning_enabled=prediction_warning_enabled,
            include_prediction_status=include_prediction_status,
            skip_drift_tracking=skip_drift_tracking,
            prediction_instance=prediction_instance,
            abort_on_error=abort_on_error,
            column_names_remapping=column_names_remapping,
            include_probabilities=include_probabilities,
            include_probabilities_classes=include_probabilities_classes,
            download_timeout=download_timeout,
            download_read_timeout=download_read_timeout,
            upload_read_timeout=upload_read_timeout,
            explanations_mode=explanations_mode,
        )

    @classmethod
    def apply_time_series_data_prep_and_score(
        cls,
        deployment: Deployment,
        intake_settings: IntakeSettings,
        timeseries_settings: TimeSeriesSettings,
        **kwargs,
    ) -> BatchPredictionJob:
        """
        Prepare the dataset with time series data prep, create new batch prediction job,
        upload the scoring dataset, and return a batch prediction job.

        The supported intake_settings are of type `localFile` or `dataset`.

        For timeseries_settings of type `forecast` the `forecast_point` must be specified.

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        .. versionadded:: v3.1

        Attributes
        ----------
        deployment : Deployment
            Deployment which will be used for scoring.

        intake_settings : dict
            A dict configuring where data is coming from. Supported options:

                - type : str, either `localFile`, `dataset`

            Note that to pass a dataset, you not only need to specify the `type` parameter
            as `dataset`, but you must also set the `dataset` parameter as a
            :py:class:`Dataset <datarobot.models.Dataset>` object.

            To score from a local file, add this parameter to the
            settings:

                - file : file-like object, string path to file or a
                  pandas.DataFrame of scoring data.

        timeseries_settings : dict
            Configuration for time-series scoring. Supported options:

            - `type` : str, must be `forecast` or `historical` (default if
              not passed is `forecast`). `forecast` mode makes predictions using
              `forecast_point`. `historical` enables bulk prediction mode which
              calculates predictions for all possible forecast points and forecast
              distances in the dataset within `predictions_start_date`/`predictions_end_date`
              range.
            - `forecast_point` : Optional[datetime.datetime], forecast point for the dataset,
              used for the forecast predictions. Must be passed if
              ``timeseries_settings.type=forecast``.
            - `predictions_start_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `predictions_end_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `relax_known_in_advance_features_check` : bool, (default `False`).
              If True, missing values in the known in advance features are allowed
              in the forecast window at the prediction time. If omitted or False,
              missing values are not allowed.

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob

        Raises
        ------
        InvalidUsageError
            If the deployment does not support time series data prep.
            If the intake type is not supported for time series data prep.
        """
        project = None
        if deployment.model:
            project_id = deployment.model.get("project_id")
            if project_id:
                project = Project.get(project_id)

        if not project or not project.query_generator_id:
            raise errors.InvalidUsageError("This deployment does not support time series data prep")

        if intake_settings["type"] == IntakeAdapters.LOCAL_FILE:
            dataset = Dataset.upload(intake_settings["file"])
        elif intake_settings["type"] == IntakeAdapters.DATASET:
            dataset = intake_settings["dataset"]
        else:
            raise errors.InvalidUsageError(
                f"Time series data prep does not support intake type {intake_settings['type']}"
            )

        prediction_dataset = Dataset.create_from_query_generator(
            project.query_generator_id, dataset.id, dataset.version_id
        )

        kwargs["intake_settings"] = {
            "type": IntakeAdapters.DATASET,
            "dataset": prediction_dataset,
        }

        return cls.score(deployment, timeseries_settings=timeseries_settings, **kwargs)

    @classmethod
    def score_to_file(
        cls,
        deployment: DeploymentType,
        intake_path,
        output_path: str,
        **kwargs,
    ):
        """
        Create new batch prediction job, upload the scoring dataset and
        download the scored CSV file concurrently.

        Will block until the entire file is scored.

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        intake_path : file-like object/string path to file/pandas.DataFrame
              Scoring data

        output_path : str
            Filename to save the result under

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        kwargs["intake_settings"] = {
            "type": IntakeAdapters.LOCAL_FILE,
            "file": intake_path,
        }

        kwargs["output_settings"] = {
            "type": IntakeAdapters.LOCAL_FILE,
            "path": output_path,
        }

        return cls.score(deployment, **kwargs)

    @classmethod
    def apply_time_series_data_prep_and_score_to_file(
        cls,
        deployment: Deployment,
        intake_path: Union[str, pd.DataFrame, io.IOBase],
        output_path: str,
        timeseries_settings: TimeSeriesSettings,
        **kwargs,
    ) -> BatchPredictionJob:
        """
        Prepare the input dataset with time series data prep. Then, create a new batch prediction
        job using the prepared AI catalog item as input and concurrently download the scored CSV
        file.

        The function call will return when the entire file is scored.

        For timeseries_settings of type `forecast` the `forecast_point` must be specified.

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        .. versionadded:: v3.1

        Attributes
        ----------
        deployment : Deployment
            The deployment which will be used for scoring.

        intake_path : file-like object/string path to file/pandas.DataFrame
              The scoring data.

        output_path : str
            The filename under which you save the result.

        timeseries_settings : dict
            Configuration for time-series scoring. Supported options:

            - `type` : str, must be `forecast` or `historical` (default if
              not passed is `forecast`). `forecast` mode makes predictions using
              `forecast_point`. `historical` enables bulk prediction mode which
              calculates predictions for all possible forecast points and forecast
              distances in the dataset within `predictions_start_date`/`predictions_end_date`
              range.
            - `forecast_point` : Optional[datetime.datetime], forecast point for the dataset,
              used for the forecast predictions. Must be passed if
              ``timeseries_settings.type=forecast``.
            - `predictions_start_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `predictions_end_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `relax_known_in_advance_features_check` : bool, (default `False`).
              If True, missing values in the known in advance features are allowed
              in the forecast window at the prediction time. If omitted or False,
              missing values are not allowed.

        Returns
        -------
        BatchPredictionJob
            Instance of `BatchPredictionJob`.

        Raises
        ------
        InvalidUsageError
            If the deployment does not support time series data prep.
        """
        kwargs["intake_settings"] = {
            "type": IntakeAdapters.LOCAL_FILE,
            "file": intake_path,
        }

        kwargs["output_settings"] = {
            "type": IntakeAdapters.LOCAL_FILE,
            "path": output_path,
        }

        return cls.apply_time_series_data_prep_and_score(
            deployment, timeseries_settings=timeseries_settings, **kwargs
        )

    @classmethod
    def score_s3(
        cls,
        deployment: DeploymentType,
        source_url: str,
        destination_url: str,
        credential=None,
        endpoint_url: Optional[str] = None,
        **kwargs,
    ):
        """
        Create new batch prediction job, with a scoring dataset from S3
        and writing the result back to S3.

        This returns immediately after the job has been created. You
        must poll for job completion using `get_status()` or
        `wait_for_completion()` (see `datarobot.models.Job`)

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        source_url : str
            The URL for the prediction dataset (e.g.: `s3://bucket/key`)

        destination_url : str
            The URL for the scored dataset (e.g.: `s3://bucket/key`)

        credential : str or Credential (optional)
            The AWS Credential object or credential id

        endpoint_url : Optional[str]
            Any non-default endpoint URL for S3 access (omit to use the default)

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        intake_settings = {
            "type": IntakeAdapters.S3,
        }

        if credential is not None:
            if isinstance(credential, Credential):
                intake_settings["credential_id"] = credential.credential_id
            else:
                intake_settings["credential_id"] = credential

        if endpoint_url is not None:
            intake_settings["endpoint_url"] = endpoint_url

        output_settings = intake_settings.copy()

        intake_settings["url"] = source_url
        output_settings["url"] = destination_url

        kwargs.update({"intake_settings": intake_settings, "output_settings": output_settings})

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_azure(
        cls,
        deployment: DeploymentType,
        source_url: str,
        destination_url: str,
        credential=None,
        **kwargs,
    ):
        """
        Create new batch prediction job, with a scoring dataset from Azure blob
        storage and writing the result back to Azure blob storage.

        This returns immediately after the job has been created. You
        must poll for job completion using `get_status()` or
        `wait_for_completion()` (see `datarobot.models.Job`).

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        source_url : str
            The URL for the prediction dataset
            (e.g.: `https://storage_account.blob.endpoint/container/blob_name`)

        destination_url : str
            The URL for the scored dataset
            (e.g.: `https://storage_account.blob.endpoint/container/blob_name`)

        credential : str or Credential (optional)
            The Azure Credential object or credential id

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        intake_settings = {
            "type": "azure",
        }

        if credential is not None:
            if isinstance(credential, Credential):
                intake_settings["credential_id"] = credential.credential_id
            else:
                intake_settings["credential_id"] = credential

        output_settings = intake_settings.copy()

        intake_settings["url"] = source_url
        output_settings["url"] = destination_url

        kwargs.update({"intake_settings": intake_settings, "output_settings": output_settings})

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_gcp(
        cls,
        deployment: DeploymentType,
        source_url: str,
        destination_url: str,
        credential=None,
        **kwargs,
    ):
        """
        Create new batch prediction job, with a scoring dataset from Google Cloud Storage
        and writing the result back to one.

        This returns immediately after the job has been created. You
        must poll for job completion using `get_status()` or
        `wait_for_completion()` (see `datarobot.models.Job`).

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        source_url : str
            The URL for the prediction dataset
            (e.g.: `http(s)://storage.googleapis.com/[bucket]/[object]`)

        destination_url : str
            The URL for the scored dataset
            (e.g.: `http(s)://storage.googleapis.com/[bucket]/[object]`)

        credential : str or Credential (optional)
            The GCP Credential object or credential id

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        intake_settings = {
            "type": IntakeAdapters.GCP,
        }

        if credential is not None:
            if isinstance(credential, Credential):
                intake_settings["credential_id"] = credential.credential_id
            else:
                intake_settings["credential_id"] = credential

        output_settings = intake_settings.copy()

        intake_settings["url"] = source_url
        output_settings["url"] = destination_url

        kwargs.update({"intake_settings": intake_settings, "output_settings": output_settings})

        return cls.score(deployment, **kwargs)

    @classmethod
    def score_from_existing(cls, batch_prediction_job_id: str) -> BatchPredictionJob:
        """Create a new batch prediction job based on the settings from a previously created one

        Attributes
        ----------
        batch_prediction_job_id: str
            ID of the previous batch prediction job

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """

        batch_job = super().get(None, batch_prediction_job_id)

        job_data = {"predictionJobId": batch_job.id}

        response = cls._client.post(url=cls._from_existing_path(), json=job_data)

        job_id = get_id_from_response(response)
        return BatchPredictionJob.get(job_id)

    @classmethod
    def score_pandas(
        cls,
        deployment: DeploymentType,
        df: pd.DataFrame,
        read_timeout: int = 660,
        **kwargs,
    ) -> Tuple[BatchPredictionJob, pd.DataFrame]:
        """
        Run a batch prediction job, with a scoring dataset from a
        pandas dataframe. The output from the prediction will be joined
        to the passed DataFrame and returned.

        Use `columnNamesRemapping` to drop or rename columns in the
        output

        This method blocks until the job has completed or raises an
        exception on errors.

        Refer to the :meth:`datarobot.models.BatchPredictionJob.score` method for details on the other
        `kwargs` parameters.

        Attributes
        ----------
        deployment : Deployment or string ID
            Deployment which will be used for scoring.

        df : pandas.DataFrame
            The dataframe to score

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictonJob
        pandas.DataFrame
            The original dataframe merged with the predictions
        """

        chunk_size = 10000
        index_label = "__DR_index__"

        class FileFromChunks(io.RawIOBase):
            """
            File-like object built from an iterator of chunks.
            """

            def __init__(self, iterator):
                self._iterator = iterator
                self._buffer = None
                self._read_bytes = 0

            def readable(self) -> bool:
                return True

            def readinto(self, b) -> int:
                read_size = len(b)
                if self._buffer:
                    if len(self._buffer) < read_size:
                        chunk = self._buffer.tobytes()
                        while len(chunk) < read_size:
                            try:
                                chunk += next(self._iterator)
                            except StopIteration:
                                break
                        chunk = memoryview(chunk)
                    else:
                        chunk = self._buffer
                else:
                    try:
                        chunk = next(self._iterator)
                    except StopIteration:
                        return 0

                    while len(chunk) < read_size:
                        try:
                            chunk += next(self._iterator)
                        except StopIteration:
                            break
                    chunk = memoryview(chunk)

                output, self._buffer = chunk[:read_size], chunk[read_size:]
                b[: len(output)] = output

                self._read_bytes += len(output)
                return len(output)

            def tell(self) -> int:
                return self._read_bytes

            def seek(self, n: int) -> None:
                if n != self.tell():
                    raise RuntimeError("Cannot seek")

        def csv_stream():
            LOG.info("Streaming DataFrame as CSV data to DataRobot")
            for chunk_no, chunked_df in enumerate(
                df[i : i + chunk_size] for i in range(0, df.shape[0], chunk_size)
            ):
                yield chunked_df.to_csv(
                    index_label=index_label, header=(chunk_no == 0), quoting=csv.QUOTE_ALL
                ).encode()

        kwargs["intake_settings"] = {
            "type": IntakeAdapters.LOCAL_FILE,
            "file": FileFromChunks(csv_stream()),
        }
        kwargs["output_settings"] = {"type": OutputAdapters.LOCAL_FILE}

        if kwargs.get("passthrough_columns_set") != "all":
            # If we're not passing everything through, ensure we're passing the index
            # used to join the results
            kwargs.setdefault("passthrough_columns", []).append(index_label)

        job = cls.score(deployment, **kwargs)

        status = job.get_status()
        LOG.info("Created Batch Prediction job ID %s", job.id)
        LOG.info("Waiting for DataRobot to start processing")

        status = job._wait_for_download(timeout=read_timeout)
        LOG.info("Job has started processing at DataRobot. Streaming results.")

        status = job.get_status()
        resp = job._client.get(status["links"]["download"], stream=True, timeout=read_timeout)
        if resp.status_code != 200:
            raise RuntimeError(
                (
                    "Got invalid response when downloading job data. "
                    "Status code: %s, Reponse: %s, Job ID: %s"
                )
                % (resp.status_code, resp.content, job.id)
            )

        scored_df = pd.read_csv(resp.raw, index_col=index_label)

        return job, df.merge(scored_df, left_index=True, right_index=True, how="inner")

    @classmethod
    def score_with_leaderboard_model(
        cls,
        model: ModelType,
        intake_settings: Optional[IntakeSettings] = None,
        output_settings: Optional[OutputSettings] = None,
        csv_settings: Optional[CsvSettings] = None,
        timeseries_settings: Optional[TimeSeriesSettings] = None,
        passthrough_columns: Optional[List[str]] = None,
        passthrough_columns_set: Optional[str] = None,
        max_explanations: Optional[int] = None,
        max_ngram_explanations: Optional[Union[int, str]] = None,
        explanation_algorithm: Optional[str] = None,
        threshold_high: Optional[float] = None,
        threshold_low: Optional[float] = None,
        prediction_threshold: Optional[float] = None,
        prediction_warning_enabled: Optional[bool] = None,
        include_prediction_status: bool = False,
        abort_on_error: bool = True,
        column_names_remapping: Optional[Dict[str, str]] = None,
        include_probabilities: bool = True,
        include_probabilities_classes: Optional[List[str]] = None,
        download_timeout: Optional[int] = 120,
        download_read_timeout: Optional[int] = 660,
        upload_read_timeout: Optional[int] = DEFAULT_TIMEOUT.UPLOAD,
        explanations_mode: Optional[PredictionExplanationsMode] = None,
    ) -> BatchPredictionJob:
        """
        Creates a new batch prediction job for a Leaderboard model by
        uploading the scoring dataset. Returns a batch prediction job.

        The default intake and output options are both `localFile`, which
        requires the caller to pass the `file` parameter and either
        download the results using the `download()` method afterwards or
        pass a path to a file where the scored data will be downloaded to.

        Attributes
        ----------
        model : Model or DatetimeModel or string ID
            Model which will be used for scoring.

        intake_settings : Optional[IntakeSettings]
            A dict configuring how data is coming from. Supported options:

                - type : str, either `localFile`, `dataset`, or `dss`.

            Note that to pass a dataset, you not only need to specify the `type` parameter
            as `dataset`, but you must also set the `dataset` parameter as a
            `dr.Dataset` object.

            To score from a local file, add the this parameter to the
            settings:

                - file : file-like object, string path to file or a
                  pandas.DataFrame of scoring data.

            To score subset of training data, use `dss` intake type and specify
            following parameters:

                - project_id : project to fetch training data from. Access to project is required.
                - partition : subset of training data to score,
                  one of ``datarobot.enums.TrainingDataSubsets``.

        output_settings : Optional[OutputSettings]
            A dict configuring how scored data is to be saved. Supported
            options:

                - type : str, `localFile`

            To save scored data to a local file, add this parameters to the
            settings:

                - path : Optional[str] The path to save the scored data
                  as a CSV file. If a path is not specified, you must download
                  the scored data yourself with `job.download()`.
                  If a path is specified, the call is blocked until the
                  job is done. If there are no other jobs currently
                  processing for the targeted prediction instance,
                  uploading, scoring, and downloading will happen in parallel
                  without waiting for a full job to complete. Otherwise,
                  it will still block, but start downloading the scored
                  data as soon as it starts generating data. This is the
                  fastest method to get predictions.

        csv_settings : Optional[CsvSettings]
            CSV intake and output settings. Supported options:

            - `delimiter` : str (optional, default `,`), fields are delimited by
              this character. Use the string `tab` to denote TSV (TAB separated values).
              Must be either a one-character string or the string `tab`.
            - `quotechar` : str (optional, default `"`), fields containing the
              delimiter must be quoted using this character.
            - `encoding` : str (optional, default `utf-8`), encoding for the CSV
              files. For example (but not limited to): `shift_jis`, `latin_1` or
              `mskanji`.

        timeseries_settings : Optional[TimeSeriesSettings]
            Configuration for time-series scoring. Supported options:

            - `type` : str, must be `forecast`, `historical` (default if
              not passed is `forecast`), or `training`. `forecast` mode makes predictions using
              `forecast_point` or rows in the dataset without target. `historical`
              enables bulk prediction mode which calculates predictions for all
              possible forecast points and forecast distances in the dataset within
              `predictions_start_date`/`predictions_end_date` range. `training` mode is
              a special case for predictions on subsets of training data. Note, that it must
              be used in conjunction with `dss` intake type only.
            - `forecast_point` : Optional[datetime.datetime], forecast point for the dataset,
              used for the forecast predictions, by default value will be inferred
              from the dataset. May be passed if ``timeseries_settings.type=forecast``.
            - `predictions_start_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `predictions_end_date` : Optional[datetime.datetime], used for historical
              predictions in order to override date from which predictions should be
              calculated. By default value will be inferred automatically from the
              dataset. May be passed if ``timeseries_settings.type=historical``.
            - `relax_known_in_advance_features_check` : bool, (default `False`).
              If True, missing values in the known in advance features are allowed
              in the forecast window at the prediction time. If omitted or False,
              missing values are not allowed.

        passthrough_columns : list[string] (optional)
            Keep these columns from the scoring dataset in the scored dataset.
            This is useful for correlating predictions with source data.

        passthrough_columns_set : Optional[str]
            To pass through every column from the scoring dataset, set this to
            `all`. Takes precedence over `passthrough_columns` if set.

        max_explanations : Optional[int]
            Compute prediction explanations for this amount of features.

        max_ngram_explanations : int or str (optional)
            Compute text explanations for this amount of ngrams. Set to `all` to return all ngram
            explanations, or set to a positive integer value to limit the amount of ngram
            explanations returned. By default no ngram explanations will be computed and returned.

        threshold_high : Optional[float]
            Only compute prediction explanations for predictions above this
            threshold. Can be combined with `threshold_low`.

        threshold_low : Optional[float]
            Only compute prediction explanations for predictions below this
            threshold. Can be combined with `threshold_high`.

        explanations_mode : PredictionExplanationsMode, optional
            Mode of prediction explanations calculation for multiclass and clustering models, if not
            specified - server default is to explain only the predicted class, identical to passing
            TopPredictionsMode(1).

        prediction_warning_enabled : Optional[bool]
            Add prediction warnings to the scored data. Currently only
            supported for regression models.

        include_prediction_status : Optional[bool]
            Include the `prediction_status` column in the output, defaults to `False`.

        abort_on_error : Optional[bool]
             Default behavior is to abort the job if too many rows fail scoring. This will free
             up resources for other jobs that may score successfully. Set to `false` to
             unconditionally score every row no matter how many errors are encountered.
             Defaults to `True`.

        column_names_remapping : Optional[Dict]
            Mapping with column renaming for output table. Defaults to `{}`.

        include_probabilities : Optional[bool]
            Flag that enables returning of all probability columns. Defaults to `True`.

        include_probabilities_classes : list (optional)
            List the subset of classes if you do not want all the classes. Defaults to `[]`.

        download_timeout : Optional[int]
            .. versionadded:: 2.22

            If using localFile output, wait this many seconds for the download to become
            available. See `download()`.

        download_read_timeout : int (optional, default 660)
            .. versionadded:: 2.22

            If using localFile output, wait this many seconds for the server to respond
            between chunks.

        upload_read_timeout: int (optional, default 600)
            .. versionadded:: 2.28

            If using localFile intake, wait this many seconds for the server to respond
            after whole dataset upload.

        prediction_threshold: Optional[float]
            .. versionadded:: 3.4.0

            Threshold is the point that sets the class boundary for a predicted value. The model
            classifies an observation below the threshold as FALSE, and an observation above the
            threshold as TRUE. In other words, DataRobot automatically assigns the positive class
            label to any prediction exceeding the threshold.
            This value can be set between 0.0 and 1.0.

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        try:
            model_id = cast("Model", model).id
        except AttributeError:
            model_id = cast(str, model)

        # Use same limitation as on UI for leaderboard predictions for now
        # (these are use case, not functional limitations - all of below should work fine):
        # * only local file or AI Catalog for intake
        # * only local file for output.
        if intake_settings and intake_settings.get("type") not in {
            IntakeAdapters.LOCAL_FILE,
            IntakeAdapters.DATASET,
            IntakeAdapters.DSS,
        }:
            raise ValueError(
                "Unsupported type parameter for intake_settings: {}".format(
                    intake_settings.get("type")
                )
            )

        if output_settings and output_settings.get("type") != OutputAdapters.LOCAL_FILE:
            raise ValueError(
                "Unsupported type parameter for output_settings: {}".format(
                    output_settings.get("type")
                )
            )

        return cls._score(
            model_id=model_id,
            intake_settings=intake_settings,
            output_settings=output_settings,
            csv_settings=csv_settings,
            timeseries_settings=timeseries_settings,
            passthrough_columns=passthrough_columns,
            passthrough_columns_set=passthrough_columns_set,
            max_explanations=max_explanations,
            max_ngram_explanations=max_ngram_explanations,
            explanation_algorithm=explanation_algorithm,
            threshold_high=threshold_high,
            threshold_low=threshold_low,
            prediction_threshold=prediction_threshold,
            prediction_warning_enabled=prediction_warning_enabled,
            include_prediction_status=include_prediction_status,
            abort_on_error=abort_on_error,
            column_names_remapping=column_names_remapping,
            include_probabilities=include_probabilities,
            include_probabilities_classes=include_probabilities_classes,
            download_timeout=download_timeout,
            download_read_timeout=download_read_timeout,
            upload_read_timeout=upload_read_timeout,
            explanations_mode=explanations_mode,
        )

    @classmethod
    def get(  # pylint: disable=arguments-differ
        cls,
        batch_prediction_job_id: str,
    ) -> BatchPredictionJob:
        """Get batch prediction job

        Attributes
        ----------
        batch_prediction_job_id: str
            ID of batch prediction job

        Returns
        -------
        BatchPredictionJob
            Instance of BatchPredictionJob
        """
        batch_job = super().get(None, batch_prediction_job_id)
        batch_job.id = batch_prediction_job_id

        return batch_job

    def download(self, fileobj, timeout: int = 120, read_timeout: int = 660) -> None:
        """Downloads the CSV result of a prediction job

        Attributes
        ----------
        fileobj: A file-like object where the CSV prediction results will be
            written to. Examples include an in-memory buffer
            (e.g., io.BytesIO) or a file on disk (opened for binary writing).

        timeout : int (optional, default 120)
            .. versionadded:: 2.22

            Seconds to wait for the download to become available.

            The download will not be available before the job has started processing.
            In case other jobs are occupying the queue, processing may not start
            immediately.

            If the timeout is reached, the job will be aborted and `RuntimeError`
            is raised.

            Set to -1 to wait infinitely.

        read_timeout : int (optional, default 660)
            .. versionadded:: 2.22

            Seconds to wait for the server to respond between chunks.
        """
        self._download(fileobj, timeout, read_timeout)

    def delete(self, ignore_404_errors: bool = False) -> None:
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        self._delete(ignore_404_errors)

    def get_status(self):
        """Get status of batch prediction job

        Returns
        -------
        BatchPredictionJob status data
            Dict with job status
        """
        return self._get_status()

    @classmethod
    def list_by_status(cls, statuses: Optional[List[str]] = None) -> List[BatchPredictionJob]:
        """Get jobs collection for specific set of statuses

        Attributes
        ----------
        statuses
            List of statuses to filter jobs ([ABORTED|COMPLETED...])
            if statuses is not provided, returns all jobs for user

        Returns
        -------
        BatchPredictionJob statuses
            List of job statuses dicts with specific statuses
        """
        if statuses is not None:
            params = {"status": statuses}
        else:
            params = {}

        return list(pagination.unpaginate(cls._jobs_path(), params, cls._client))

    def _make_result_from_location(self, location: str, *args, **kwargs):
        server_data = self._client.get(location)
        if server_data.status_code == 200:
            return server_data.content

    @classmethod
    def _data_and_completed_url_for_job(cls, url: str):
        response = cls._client.get(url, allow_redirects=False)

        if response.status_code == 200:
            data = response.json()
            if "download" in data["links"]:
                completed_url = data["links"]["download"]
            else:
                completed_url = data["links"]["self"]
            return data, completed_url
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise errors.AsyncFailureError(e_msg.format(response.status_code))

    def __repr__(self) -> str:
        return f"BatchPredictionJob({self.job_type}, '{self.id}', status={self.status})"


class BatchPredictionJobDefinition(APIObject):  # pylint: disable=missing-class-docstring
    _path = "batchPredictionJobDefinitions/"

    _user = t.Dict(
        {
            t.Key("username"): String(),
            t.Key("full_name", optional=True): String(),
            t.Key("user_id"): String(),
        }
    ).allow_extra("*")

    _schedule = t.Dict(
        {
            t.Key("day_of_week"): t.List(t.Or(String, Int)),
            t.Key("month"): t.List(t.Or(String, Int)),
            t.Key("hour"): t.List(t.Or(String, Int)),
            t.Key("minute"): t.List(t.Or(String, Int)),
            t.Key("day_of_month"): t.List(t.Or(String, Int)),
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String,
            t.Key("name"): String,
            t.Key("enabled"): t.Bool(),
            t.Key("schedule", optional=True): _schedule,
            t.Key("batch_prediction_job"): BatchPredictionJob._job_spec,
            t.Key("created"): String(),
            t.Key("updated"): String(),
            t.Key("created_by"): _user,
            t.Key("updated_by"): _user,
            t.Key("last_failed_run_time", optional=True): String(),
            t.Key("last_successful_run_time", optional=True): String(),
            t.Key("last_successful_run_time", optional=True): String(),
            t.Key("last_scheduled_run_time", optional=True): String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        enabled: Optional[bool] = None,
        schedule: Optional[Schedule] = None,
        batch_prediction_job=None,
        created: Optional[str] = None,
        updated: Optional[str] = None,
        created_by=None,
        updated_by=None,
        last_failed_run_time: Optional[str] = None,
        last_successful_run_time: Optional[str] = None,
        last_started_job_status: Optional[str] = None,
        last_scheduled_run_time: Optional[str] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.enabled = enabled
        self.schedule = schedule
        self.batch_prediction_job = batch_prediction_job

        self.created = created
        self.updated = updated
        self.created_by = created_by
        self.updated_by = updated_by

        self.last_failed_run_time = last_failed_run_time
        self.last_successful_run_time = last_successful_run_time
        self.last_started_job_status = last_started_job_status
        self.last_scheduled_run_time = last_scheduled_run_time

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"

    def __eq__(self, other) -> bool:
        return isinstance(other, self.__class__) and self.id == other.id

    @classmethod
    def get(cls, batch_prediction_job_definition_id: str) -> BatchPredictionJobDefinition:
        """Get batch prediction job definition

        Attributes
        ----------
        batch_prediction_job_definition_id: str
            ID of batch prediction job definition

        Returns
        -------
        BatchPredictionJobDefinition
            Instance of BatchPredictionJobDefinition

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchPredictionJobDefinition.get('5a8ac9ab07a57a0001be501f')
            >>> definition
            BatchPredictionJobDefinition(60912e09fd1f04e832a575c1)
        """

        return cls.from_location(f"{cls._path}{batch_prediction_job_definition_id}/")

    @classmethod
    def list(
        cls,
        search_name: Optional[str] = None,
        deployment_id: Optional[str] = None,
        limit: Union[int, MissingType] = MISSING,
        offset: int = 0,
    ) -> List[BatchPredictionJobDefinition]:
        """
        Get job all definitions

        Parameters
        ----------
        search_name : Optional[str]
            String for filtering job definitions
            Job definitions that contain the string in name will be returned.
            If not specified, all available job definitions will be returned.
        deployment_id: str
            The ID of the deployment record belongs to.
        limit: Optional[int]
            0 by default. At most this many results are returned.
        offset: Optional[int]
            This many results will be skipped.

        Returns
        -------
        List[BatchPredictionJobDefinition]
            List of job definitions the user has access to see

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchPredictionJobDefinition.list()
            >>> definition
            [
                BatchPredictionJobDefinition(60912e09fd1f04e832a575c1),
                BatchPredictionJobDefinition(6086ba053f3ef731e81af3ca)
            ]
        """
        if limit is MISSING:
            deprecation_warning(
                "BatchPredictionJobDefinition.list",
                deprecated_since_version="v3.4",
                will_remove_version="v3.6",
                message="BatchPredictionJobDefinition.list will no longer return all job definitions "
                "after version 3.6 is released. To preserve current behavior please pass limit=0.",
            )

            limit = 0  # deprecated behaviour: return everything

        params = {}
        if search_name:
            params["searchName"] = search_name
        if deployment_id:
            params["deploymentId"] = deployment_id

        if limit == 0:  # unlimited results
            r_data = pagination.unpaginate(cls._path, params, cls._client)
        else:
            params["offset"] = offset
            params["limit"] = limit
            response = cls._client.get(cls._path, params=params).json()
            r_data = response["data"]

        records = cast(
            List["BatchPredictionJobDefinition"], [cls.from_server_data(item) for item in r_data]
        )
        return records

    @classmethod
    def create(
        cls,
        enabled: bool,
        batch_prediction_job,
        name: Optional[str] = None,
        schedule: Optional[Schedule] = None,
    ) -> BatchPredictionJobDefinition:
        """
        Creates a new batch prediction job definition to be run either at scheduled interval or as
        a manual run.

        Attributes
        ----------
        enabled : bool (default False)
            Whether or not the definition should be active on a scheduled basis. If True,
            `schedule` is required.

        batch_prediction_job: dict
            The job specifications for your batch prediction job.
            It requires the same job input parameters as used with
            :func:`~BatchPredictionJob.score`, only it will not initialize a job scoring,
            only store it as a definition for later use.

        name : Optional[str]
            The name you want your job to be identified with. Must be unique across the
            organization's existing jobs.
            If you don't supply a name, a random one will be generated for you.

        schedule : Optional[Dict]
            The ``schedule`` payload defines at what intervals the job should run, which can be
            combined in various ways to construct complex scheduling terms if needed. In all of
            the elements in the objects, you can supply either an asterisk ``["*"]`` denoting
            "every" time denomination or an array of integers (e.g. ``[1, 2, 3]``) to define
            a specific interval.

            The ``schedule`` payload is split up in the following items:

            **Minute:**

            The minute(s) of the day that the job will run. Allowed values are either ``["*"]``
            meaning every minute of the day or ``[0 ... 59]``

            **Hour:**
            The hour(s) of the day that the job will run. Allowed values are either ``["*"]``
            meaning every hour of the day or ``[0 ... 23]``.

            **Day of Month:**
            The date(s) of the month that the job will run. Allowed values are either
            ``[1 ... 31]`` or ``["*"]`` for all days of the month. This field is additive with
            ``dayOfWeek``, meaning the job will run both on the date(s) defined in this field
            and the day specified by ``dayOfWeek`` (for example, dates 1st, 2nd, 3rd, plus every
            Tuesday). If ``dayOfMonth`` is set to ``["*"]`` and ``dayOfWeek`` is defined,
            the scheduler will trigger on every day of the month that matches ``dayOfWeek``
            (for example, Tuesday the 2nd, 9th, 16th, 23rd, 30th).
            Invalid dates such as February 31st are ignored.

            **Month:**
            The month(s) of the year that the job will run. Allowed values are either
            ``[1 ... 12]`` or ``["*"]`` for all months of the year. Strings, either
            3-letter abbreviations or the full name of the month, can be used
            interchangeably (e.g., "jan" or "october").
            Months that are not compatible with ``dayOfMonth`` are ignored, for example
            ``{"dayOfMonth": [31], "month":["feb"]}``

            **Day of Week:**
            The day(s) of the week that the job will run. Allowed values are ``[0 .. 6]``,
            where (Sunday=0), or ``["*"]``, for all days of the week. Strings, either 3-letter
            abbreviations or the full name of the day, can be used interchangeably
            (e.g., "sunday", "Sunday", "sun", or "Sun", all map to ``[0]``.
            This field is additive with ``dayOfMonth``, meaning the job will run both on the
            date specified by ``dayOfMonth`` and the day defined in this field.

        Returns
        -------
        BatchPredictionJobDefinition
            Instance of BatchPredictionJobDefinition

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> job_spec = {
            ...    "num_concurrent": 4,
            ...    "deployment_id": "foobar",
            ...    "intake_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...    "output_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...}
            >>> schedule = {
            ...    "day_of_week": [
            ...        1
            ...    ],
            ...    "month": [
            ...        "*"
            ...    ],
            ...    "hour": [
            ...        16
            ...    ],
            ...    "minute": [
            ...        0
            ...    ],
            ...    "day_of_month": [
            ...        1
            ...    ]
            ...}
            >>> definition = BatchPredictionJobDefinition.create(
            ...    enabled=False,
            ...    batch_prediction_job=job_spec,
            ...    name="some_definition_name",
            ...    schedule=schedule
            ... )
            >>> definition
            BatchPredictionJobDefinition(60912e09fd1f04e832a575c1)
        """

        BatchPredictionJob._job_spec.check(batch_prediction_job)

        job_spec = to_api(batch_prediction_job)

        payload = {
            "name": name,
            "enabled": enabled,
        }

        if schedule:
            payload["schedule"] = to_api(schedule)

        payload.update(**job_spec)

        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def update(
        self,
        enabled: bool,
        batch_prediction_job=None,
        name: Optional[str] = None,
        schedule: Optional[Schedule] = None,
    ) -> BatchPredictionJobDefinition:
        """
        Updates a job definition with the changed specs.

        Takes the same input as :func:`~BatchPredictionJobDefinition.create`

        Attributes
        ----------
        enabled : bool (default False)
            Same as ``enabled`` in :func:`~BatchPredictionJobDefinition.create`.

        batch_prediction_job: dict
            Same as ``batch_prediction_job`` in :func:`~BatchPredictionJobDefinition.create`.

        name : Optional[str]
            Same as ``name`` in :func:`~BatchPredictionJobDefinition.create`.

        schedule : dict
            Same as ``schedule`` in :func:`~BatchPredictionJobDefinition.create`.

        Returns
        -------
        BatchPredictionJobDefinition
            Instance of the updated BatchPredictionJobDefinition

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> job_spec = {
            ...    "num_concurrent": 5,
            ...    "deployment_id": "foobar_new",
            ...    "intake_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...    "output_settings": {
            ...        "url": "s3://foobar/123",
            ...        "type": "s3",
            ...        "format": "csv"
            ...    },
            ...}
            >>> schedule = {
            ...    "day_of_week": [
            ...        1
            ...    ],
            ...    "month": [
            ...        "*"
            ...    ],
            ...    "hour": [
            ...        "*"
            ...    ],
            ...    "minute": [
            ...        30, 59
            ...    ],
            ...    "day_of_month": [
            ...        1, 2, 6
            ...    ]
            ...}
            >>> definition = BatchPredictionJobDefinition.create(
            ...    enabled=False,
            ...    batch_prediction_job=job_spec,
            ...    name="updated_definition_name",
            ...    schedule=schedule
            ... )
            >>> definition
            BatchPredictionJobDefinition(60912e09fd1f04e832a575c1)
        """
        payload = {
            "enabled": enabled,
        }

        if name:
            payload["name"] = name

        if schedule:
            payload["schedule"] = to_api(schedule)

        if batch_prediction_job:
            BatchPredictionJob._job_spec.check(batch_prediction_job)
            job_spec = to_api(batch_prediction_job)
            payload.update(**job_spec)

        return self.from_server_data(
            self._client.patch(f"{self._path}{self.id}", data=payload).json()
        )

    def run_on_schedule(self, schedule: Schedule) -> BatchPredictionJobDefinition:
        """
        Sets the run schedule of an already created job definition.

        If the job was previously not enabled, this will also set the job to enabled.

        Attributes
        ----------
        schedule : dict
            Same as ``schedule`` in :func:`~BatchPredictionJobDefinition.create`.

        Returns
        -------
        BatchPredictionJobDefinition
            Instance of the updated BatchPredictionJobDefinition with the new / updated schedule.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchPredictionJobDefinition.create('...')
            >>> schedule = {
            ...    "day_of_week": [
            ...        1
            ...    ],
            ...    "month": [
            ...        "*"
            ...    ],
            ...    "hour": [
            ...        "*"
            ...    ],
            ...    "minute": [
            ...        30, 59
            ...    ],
            ...    "day_of_month": [
            ...        1, 2, 6
            ...    ]
            ...}
            >>> definition.run_on_schedule(schedule)
            BatchPredictionJobDefinition(60912e09fd1f04e832a575c1)
        """

        payload = {
            "enabled": True,
            "schedule": to_api(schedule),
        }

        return self.from_server_data(
            self._client.patch(f"{self._path}{self.id}", data=payload).json()
        )

    def run_once(self) -> BatchPredictionJob:
        """
        Manually submits a batch prediction job to the queue, based off of an already
        created job definition.

        Returns
        -------
        BatchPredictionJob
          Instance of BatchPredictionJob

        Examples
        --------
        .. code-block:: python

          >>> import datarobot as dr
          >>> definition = dr.BatchPredictionJobDefinition.create('...')
          >>> job = definition.run_once()
          >>> job.wait_for_completion()
        """

        definition = self.from_location(f"{self._path}{self.id}/")

        payload = {"jobDefinitionId": definition.id}

        response = self._client.post(
            f"{BatchPredictionJob._jobs_path()}fromJobDefinition/", data=payload
        ).json()

        job_id = response["id"]
        return BatchPredictionJob.get(job_id)

    def delete(self) -> None:
        """
        Deletes the job definition and disables any future schedules of this job if any.
        If a scheduled job is currently running, this will not be cancelled.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> definition = dr.BatchPredictionJobDefinition.get('5a8ac9ab07a57a0001be501f')
            >>> definition.delete()
        """

        self._client.delete(f"{self._path}{self.id}/")



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from __future__ import annotations

from typing import Any, Dict, Union

import trafaret as t

from datarobot.enums import enum_to_list, NemoLLMType
from datarobot.models.api_object import APIObject
from datarobot.models.genai.comparison_chat import get_entity_id
from datarobot.models.genai.playground import Playground
from datarobot.models.genai.playground_moderation_configuration import (
    moderation_configuration_without_id,
    ModerationConfigurationWithoutId,
)

nemo_llm_configuration = t.Dict(
    {
        t.Key("llm_type"): t.Enum(*enum_to_list(NemoLLMType)),
        t.Key("openai_api_key_id"): t.String,
        t.Key("openai_api_base", optional=True): t.Or(t.String, t.Null),
        t.Key("openai_api_deployment_id", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")

nemo_file_contents = t.Dict(
    {
        t.Key("config_yaml_file_contents"): t.String,
        t.Key("flow_definition_file_contents"): t.String,
        t.Key("prompts_file_contents"): t.String,
    }
).ignore_extra("*")

nemo_file_contents_for_response = t.Dict(
    {
        t.Key("actions_file_contents"): t.String,
        t.Key("config_yaml_file_contents"): t.String,
        t.Key("flow_definition_file_contents"): t.String,
        t.Key("prompts_file_contents"): t.String,
    }
).ignore_extra("*")

nemo_configuration = t.Dict(
    {
        t.Key("prompt_pipeline_metric_name", optional=True): t.Or(t.String, t.Null),
        t.Key("prompt_pipeline_files", optional=True): t.Or(
            nemo_file_contents_for_response, t.Null
        ),
        t.Key("prompt_llm_configuration", optional=True): t.Or(nemo_llm_configuration, t.Null),
        t.Key("prompt_moderation_configuration", optional=True): t.Or(
            moderation_configuration_without_id, t.Null
        ),
        t.Key("prompt_pipeline_template_id", optional=True): t.Or(t.String, t.Null),
        t.Key("response_pipeline_metric_name", optional=True): t.Or(t.String, t.Null),
        t.Key("response_pipeline_files", optional=True): t.Or(
            nemo_file_contents_for_response, t.Null
        ),
        t.Key("response_llm_configuration", optional=True): t.Or(nemo_llm_configuration, t.Null),
        t.Key("response_moderation_configuration", optional=True): t.Or(
            moderation_configuration_without_id, t.Null
        ),
        t.Key("response_pipeline_template_id", optional=True): t.Or(t.String, t.Null),
        t.Key("blocked_terms_file_contents"): t.String,
    }
).ignore_extra("*")


class NemoLLMConfiguration(APIObject):
    """
    Configuration for the LLM model to be used for Nemo Pipeline.

    Attributes
    ----------
    llm_type : NemoLLMType
        The type of LLM model.
    openai_api_key_id : str
        The ID of the credential in the Datarobot credential manager for the API token.
    openai_api_base : Optional[str]
        The base URL for Azure OpenAI API.
    openai_api_deployment_id : Optional[str]
        The deployment ID for Azure OpenAI API.

    """

    _converter = nemo_llm_configuration

    def __init__(
        self,
        llm_type: NemoLLMType,
        openai_api_key_id: str,
        openai_api_base: Union[str, None] = None,
        openai_api_deployment_id: Union[str, None] = None,
    ):
        self.llm_type = llm_type
        self.openai_api_key_id = openai_api_key_id
        self.openai_api_base = openai_api_base
        self.openai_api_deployment_id = openai_api_deployment_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"

    def to_dict(self) -> dict[str, Any]:
        return {
            "llm_type": self.llm_type,
            "openai_api_key_id": self.openai_api_key_id,
            "openai_api_base": self.openai_api_base,
            "openai_api_deployment_id": self.openai_api_deployment_id,
        }


class NemoFileContents(APIObject):
    """
    String representation of the contents of files used in the Nemo Pipeline.
    This is the base user configurable values.

    Attributes
    ----------
    config_yaml_file_contents : str
        The contents of the config YAML file.
    flow_definition_file_contents : str
        The contents of the flow definition fle.
    prompts_file_contents : str
        The contents of the prompts file.
    """

    _converter = nemo_file_contents

    def __init__(
        self,
        config_yaml_file_contents: str,
        flow_definition_file_contents: str,
        prompts_file_contents: str,
    ):
        self.config_yaml_file_contents = config_yaml_file_contents
        self.flow_definition_file_contents = flow_definition_file_contents
        self.prompts_file_contents = prompts_file_contents

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"

    def to_dict(self) -> dict[str, str]:
        return {
            "config_yaml_file_contents": self.config_yaml_file_contents,
            "flow_definition_file_contents": self.flow_definition_file_contents,
            "prompts_file_contents": self.prompts_file_contents,
        }


class NemoFileContentsResponse(APIObject):
    """
    String representation of the contents of files used in the Nemo Pipeline.
    This is the files actually used by nemo, with the included actions python file.

    Attributes
    ----------
    actions_file_contents : str
        The contents of the actions python file.
    config_yaml_file_contents : str
        The contents of the config YAML file.
    flow_definition_file_contents : str
        The contents of the flow definition fle.
    prompts_file_contents : str
        The contents of the prompts file.
    """

    _converter = nemo_file_contents_for_response

    def __init__(
        self,
        actions_file_contents: str,
        config_yaml_file_contents: str,
        flow_definition_file_contents: str,
        prompts_file_contents: str,
    ):
        self.actions_file_contents = actions_file_contents
        self.config_yaml_file_contents = config_yaml_file_contents
        self.flow_definition_file_contents = flow_definition_file_contents
        self.prompts_file_contents = prompts_file_contents

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"

    def to_dict(self) -> dict[str, str]:
        return {
            "actions_file_contents": self.actions_file_contents,
            "config_yaml_file_contents": self.config_yaml_file_contents,
            "flow_definition_file_contents": self.flow_definition_file_contents,
            "prompts_file_contents": self.prompts_file_contents,
        }


class NemoConfiguration(APIObject):
    """
    Configuration for the Nemo Pipeline.

    Attributes
    ----------
    prompt_pipeline_metric_name : Optional[str]
        The name of the metric for the prompt pipeline.
    prompt_pipeline_files : NemoFileContentsResponse, optional
        The files used in the prompt pipeline.
    prompt_llm_configuration : NemoLLMConfiguration, optional
        The LLM configuration for the prompt pipeline.
    prompt_moderation_configuration : ModerationConfigurationWithoutID, optional
        The moderation configuration for the prompt pipeline.
    prompt_pipeline_template_id : Optional[str]
        The ID of the prompt pipeline template. This parameter defines the actions.py file.
    response_pipeline_metric_name : Optional[str]
        The name of the metric for the response pipeline.
    response_pipeline_files : NemoFileContentsResponse, optional
        The files used in the response pipeline.
    response_llm_configuration : NemoLLMConfiguration, optional
        The LLM configuration for the response pipeline.
    response_moderation_configuration : ModerationConfigurationWithoutID, optional
        The moderation configuration for the response pipeline.
    response_pipeline_template_id : Optional[str]
        The ID of the response pipeline template. This parameter defines the actions.py file.
    blocked_terms_file_contents : str
        The contents of the blocked terms file.  This is shared between the prompt and response pipelines.
    """

    _path = "api/v2/genai/playgrounds/{}/nemoConfiguration/"

    _converter = nemo_configuration
    schema = _converter

    def __init__(
        self,
        blocked_terms_file_contents: str,
        prompt_pipeline_metric_name: Union[str, None] = None,
        prompt_pipeline_files: Union[Dict[str, str], None] = None,
        prompt_llm_configuration: Union[Dict[str, Any], None] = None,
        prompt_moderation_configuration: Union[Dict[str, Any], None] = None,
        prompt_pipeline_template_id: Union[str, None] = None,
        response_pipeline_metric_name: Union[str, None] = None,
        response_pipeline_files: Union[Dict[str, str], None] = None,
        response_llm_configuration: Union[Dict[str, Any], None] = None,
        response_moderation_configuration: Union[Dict[str, Any], None] = None,
        response_pipeline_template_id: Union[str, None] = None,
    ):
        def _get_obj(value: Union[Dict[str, Any], None], obj: Any) -> Any:
            return obj.from_server_data(value) if value is not None else None

        self.prompt_pipeline_metric_name = prompt_pipeline_metric_name
        self.prompt_pipeline_files = _get_obj(prompt_pipeline_files, NemoFileContentsResponse)
        self.prompt_llm_configuration = _get_obj(prompt_llm_configuration, NemoLLMConfiguration)
        self.prompt_moderation_configuration = _get_obj(
            prompt_moderation_configuration, ModerationConfigurationWithoutId
        )
        self.prompt_pipeline_template_id = prompt_pipeline_template_id
        self.response_pipeline_metric_name = response_pipeline_metric_name
        self.response_pipeline_files = _get_obj(response_pipeline_files, NemoFileContentsResponse)
        self.response_llm_configuration = _get_obj(response_llm_configuration, NemoLLMConfiguration)
        self.response_moderation_configuration = _get_obj(
            response_moderation_configuration, ModerationConfigurationWithoutId
        )
        self.response_pipeline_template_id = response_pipeline_template_id
        self.blocked_terms_file_contents = blocked_terms_file_contents

    @classmethod
    def get(cls, playground: Union[str, Playground]) -> NemoConfiguration:
        """
        Get the Nemo configuration for a playground.

        Parameters
        ----------
        playground: str or Playground
            The playground to get the configuration for

        Returns
        -------
        NemoConfiguration
            The Nemo configuration for the playground.

        """
        playground_id = get_entity_id(playground)
        url = f"{cls._client.domain}/{cls._path.format(playground_id)}"
        response = cls._client.get(url)
        return cls.from_server_data(response.json())

    @classmethod
    def upsert(
        cls,
        playground: Union[str, Playground],
        blocked_terms_file_contents: str,
        prompt_pipeline_metric_name: Union[str, None] = None,
        prompt_pipeline_files: Union[NemoFileContents, None] = None,
        prompt_llm_configuration: Union[NemoLLMConfiguration, None] = None,
        prompt_moderation_configuration: Union[ModerationConfigurationWithoutId, None] = None,
        prompt_pipeline_template_id: Union[str, None] = None,
        response_pipeline_metric_name: Union[str, None] = None,
        response_pipeline_files: Union[NemoFileContents, None] = None,
        response_llm_configuration: Union[NemoLLMConfiguration, None] = None,
        response_moderation_configuration: Union[ModerationConfigurationWithoutId, None] = None,
        response_pipeline_template_id: Union[str, None] = None,
    ) -> NemoConfiguration:
        """
        Create or update the nemo configuration for a playground.

        Parameters
        ----------
        playground: str or Playground
            The playground for the configuration
        blocked_terms_file_contents: str
            The contents of the blocked terms file.
        prompt_pipeline_metric_name: Optional[str]
            The name of the metric for the prompt pipeline.
        prompt_pipeline_files: NemoFileContents, optional
            The files used in the prompt pipeline.
        prompt_llm_configuration: NemoLLMConfiguration, optional
            The LLM configuration for the prompt pipeline.
        prompt_moderation_configuration: ModerationConfigurationWithoutID, optional
            The moderation configuration for the prompt pipeline.
        prompt_pipeline_template_id: Optional[str]
            The ID of the prompt pipeline template, this will define the action.py file.
        response_pipeline_metric_name: Optional[str]
            The name of the metric for the response pipeline.
        response_pipeline_files: NemoFileContents, optional
            The files used in the response pipeline.
        response_llm_configuration: NemoLLMConfiguration, optional
            The LLM configuration for the response pipeline.
        response_moderation_configuration: ModerationConfigurationWithoutID, optional
            The moderation configuration for the response pipeline.
        response_pipeline_template_id: Optional[str]
            The ID of the response pipeline template, this will define the action.py file.

        Returns
        -------
        NemoConfiguration
            The Nemo configuration for the playground.
        """

        def _get_dict(
            value: Union[
                NemoFileContents, NemoLLMConfiguration, ModerationConfigurationWithoutId, None
            ]
        ) -> Union[Dict[str, Any], None]:
            return value.to_dict() if value is not None else None

        playground_id = get_entity_id(playground)
        url = f"{cls._client.domain}/{cls._path.format(playground_id)}"
        data = {
            "blocked_terms_file_contents": blocked_terms_file_contents,
            "prompt_pipeline_metric_name": prompt_pipeline_metric_name,
            "prompt_pipeline_files": _get_dict(prompt_pipeline_files),
            "prompt_llm_configuration": _get_dict(prompt_llm_configuration),
            "prompt_moderation_configuration": _get_dict(prompt_moderation_configuration),
            "prompt_pipeline_template_id": prompt_pipeline_template_id,
            "response_pipeline_metric_name": response_pipeline_metric_name,
            "response_pipeline_files": _get_dict(response_pipeline_files),
            "response_llm_configuration": _get_dict(response_llm_configuration),
            "response_moderation_configuration": _get_dict(response_moderation_configuration),
            "response_pipeline_template_id": response_pipeline_template_id,
        }
        response = cls._client.post(url, json=data)
        return cls.from_server_data(response.json())



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, List

import trafaret as t

from datarobot.insights.base import BaseInsight


class LiftChart(BaseInsight):
    """Class for Lift Chart calculations. Use the standard methods of BaseInsight to compute
    and retrieve: compute, create, list, get.

    Usage example:

        ```python
        >>> from datarobot.insights import LiftChart
        >>> LiftChart.compute("67643b2d87bb4954d7917323", data_slice_id="6764389b4bdd48581485a58b")
        <datarobot.models.status_check_job.StatusCheckJob object at 0x7fa7f5ba20b0>
        >>> LiftChart.list("67643b2d87bb4954d7917323")
        [<datarobot.insights.lift_chart.LiftChart object at 0x7fe242eeaa10>, ... ]
        >>> LiftChart.get("67643b2d87bb4954d7917323", data_slice_id="6764389b4bdd48581485a58b").bins
        [{'actual': 0.4, 'predicted': 0.22727272727272724, 'bin_weight': 5.0}, ... ]
        ```
    """

    LIFT_CHART_BINS = t.Dict(
        {t.Key("actual"): t.Float(), t.Key("predicted"): t.Float(), t.Key("bin_weight"): t.Float()}
    ).ignore_extra("*")

    INSIGHT_NAME = "liftChart"
    INSIGHT_DATA = {t.Key("bins"): t.List(LIFT_CHART_BINS)}

    @property
    def bins(self) -> List[Dict[str, Any]]:
        """Lift chart bins."""
        return cast(List[Dict[str, Any]], self.data["bins"])
